import typing as ta

import anyio

from ... import lang


T = ta.TypeVar('T')


##


class Once:
    def __init__(self) -> None:
        super().__init__()
        self._done = False
        self._lock = anyio.Lock()

    async def do(self, fn: ta.Callable[[], ta.Awaitable[None]]) -> bool:
        if self._done:
            return False
        async with self._lock:
            if self._done:
                return False  # type: ignore
            try:
                await fn()
            finally:
                self._done = True
            return True


##


class Lazy(ta.Generic[T]):
    def __init__(self) -> None:
        super().__init__()
        self._once = Once()
        self._v: lang.Maybe[T] = lang.empty()

    def peek(self) -> lang.Maybe[T]:
        return self._v

    def set(self, v: T) -> None:
        self._v = lang.just(v)

    async def get(self, fn: ta.Callable[[], ta.Awaitable[T]]) -> T:
        async def do():
            self._v = lang.just(await fn())
        await self._once.do(do)
        return self._v.must()


class LazyFn(ta.Generic[T]):
    def __init__(self, fn: ta.Callable[[], ta.Awaitable[T]]) -> None:
        super().__init__()
        self._fn = fn
        self._once = Once()
        self._v: lang.Maybe[T] = lang.empty()

    def peek(self) -> lang.Maybe[T]:
        return self._v

    def set(self, v: T) -> None:
        self._v = lang.just(v)

    async def get(self) -> T:
        async def do():
            self._v = lang.just(await self._fn())
        await self._once.do(do)
        return self._v.must()



import typing as ta

from ..base import MarshalContext
from ..base import Marshaler
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..values import Value


##


class NopMarshalerUnmarshaler(Marshaler, Unmarshaler):
    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        return o  # noqa

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return v


NOP_MARSHALER_UNMARSHALER = NopMarshalerUnmarshaler()



import typing as ta

from .base import MarshalerFactory
from .base import MultiMarshalerFactory
from .base import MultiUnmarshalerFactory
from .base import RecursiveMarshalerFactory
from .base import RecursiveUnmarshalerFactory
from .base import TypeCacheMarshalerFactory
from .base import TypeCacheUnmarshalerFactory
from .base import UnmarshalerFactory
from .composite.iterables import IterableMarshalerFactory
from .composite.iterables import IterableUnmarshalerFactory
from .composite.literals import LiteralMarshalerFactory
from .composite.literals import LiteralUnmarshalerFactory
from .composite.mappings import MappingMarshalerFactory
from .composite.mappings import MappingUnmarshalerFactory
from .composite.maybes import MaybeMarshalerFactory
from .composite.maybes import MaybeUnmarshalerFactory
from .composite.newtypes import NewtypeMarshalerFactory
from .composite.newtypes import NewtypeUnmarshalerFactory
from .composite.optionals import OptionalMarshalerFactory
from .composite.optionals import OptionalUnmarshalerFactory
from .composite.special import SequenceNotStrMarshalerFactory
from .composite.special import SequenceNotStrUnmarshalerFactory
from .objects.dataclasses import DataclassMarshalerFactory
from .objects.dataclasses import DataclassUnmarshalerFactory
from .objects.namedtuples import NamedtupleMarshalerFactory
from .objects.namedtuples import NamedtupleUnmarshalerFactory
from .polymorphism.unions import PrimitiveUnionMarshalerFactory
from .polymorphism.unions import PrimitiveUnionUnmarshalerFactory
from .singular.datetimes import DATETIME_MARSHALER_FACTORY
from .singular.datetimes import DATETIME_UNMARSHALER_FACTORY
from .singular.enums import EnumMarshalerFactory
from .singular.enums import EnumUnmarshalerFactory
from .singular.numbers import NUMBERS_MARSHALER_FACTORY
from .singular.numbers import NUMBERS_UNMARSHALER_FACTORY
from .singular.primitives import PRIMITIVE_MARSHALER_FACTORY
from .singular.primitives import PRIMITIVE_UNMARSHALER_FACTORY
from .singular.uuids import UUID_MARSHALER_FACTORY
from .singular.uuids import UUID_UNMARSHALER_FACTORY
from .trivial.any import ANY_MARSHALER_FACTORY
from .trivial.any import ANY_UNMARSHALER_FACTORY


##


STANDARD_MARSHALER_FACTORIES: list[MarshalerFactory] = [
    PRIMITIVE_MARSHALER_FACTORY,
    NewtypeMarshalerFactory(),
    OptionalMarshalerFactory(),
    PrimitiveUnionMarshalerFactory(),
    DataclassMarshalerFactory(),
    NamedtupleMarshalerFactory(),
    EnumMarshalerFactory(),
    LiteralMarshalerFactory(),
    NUMBERS_MARSHALER_FACTORY,
    UUID_MARSHALER_FACTORY,
    DATETIME_MARSHALER_FACTORY,
    MaybeMarshalerFactory(),
    MappingMarshalerFactory(),
    SequenceNotStrMarshalerFactory(),
    IterableMarshalerFactory(),
    ANY_MARSHALER_FACTORY,
]


def new_standard_marshaler_factory(
        *,
        first: ta.Iterable[MarshalerFactory] | None = None,
        last: ta.Iterable[MarshalerFactory] | None = None,
) -> MarshalerFactory:
    return TypeCacheMarshalerFactory(
        RecursiveMarshalerFactory(
            MultiMarshalerFactory([
                *(first if first is not None else []),
                *STANDARD_MARSHALER_FACTORIES,
                *(last if last is not None else []),
            ]),
        ),
    )


##


STANDARD_UNMARSHALER_FACTORIES: list[UnmarshalerFactory] = [
    PRIMITIVE_UNMARSHALER_FACTORY,
    NewtypeUnmarshalerFactory(),
    OptionalUnmarshalerFactory(),
    PrimitiveUnionUnmarshalerFactory(),
    DataclassUnmarshalerFactory(),
    NamedtupleUnmarshalerFactory(),
    EnumUnmarshalerFactory(),
    LiteralUnmarshalerFactory(),
    NUMBERS_UNMARSHALER_FACTORY,
    UUID_UNMARSHALER_FACTORY,
    DATETIME_UNMARSHALER_FACTORY,
    MaybeUnmarshalerFactory(),
    MappingUnmarshalerFactory(),
    SequenceNotStrUnmarshalerFactory(),
    IterableUnmarshalerFactory(),
    ANY_UNMARSHALER_FACTORY,
]


def new_standard_unmarshaler_factory(
        *,
        first: ta.Iterable[UnmarshalerFactory] | None = None,
        last: ta.Iterable[UnmarshalerFactory] | None = None,
) -> UnmarshalerFactory:
    return TypeCacheUnmarshalerFactory(
        RecursiveUnmarshalerFactory(
            MultiUnmarshalerFactory([
                *(first if first is not None else []),
                *STANDARD_UNMARSHALER_FACTORIES,
                *(last if last is not None else []),
            ]),
        ),
    )


##


def install_standard_factories(
        *factories: MarshalerFactory | UnmarshalerFactory,
) -> None:
    for f in factories:
        k = False

        if isinstance(f, MarshalerFactory):
            STANDARD_MARSHALER_FACTORIES[0:0] = [f]
            k = True

        if isinstance(f, UnmarshalerFactory):
            STANDARD_UNMARSHALER_FACTORIES[0:0] = [f]
            k = True

        if not k:
            raise TypeError(f)



"""
TODO:
 - F_SETLK mode
"""
import os
import typing as ta

from ... import lang
from ...argparse import all as ap
from ..pidfiles.pidfile import Pidfile
from ..pidfiles.pinning import PidfilePinner
from ..signals import parse_signal


##


class Cli(ap.Cli):
    _PIDFILE_ARGS: ta.ClassVar[ta.Sequence[ap.Arg]] = [
        ap.arg('pid-file'),
        ap.arg('--create', action='store_true'),
    ]

    def _pidfile_args(self) -> lang.Args:
        return lang.Args(
            self.args.pid_file,
            inheritable=False,
            no_create=not self._args.create,
        )

    def _args_pidfile(self) -> Pidfile:
        return self._pidfile_args()(Pidfile)

    #

    @ap.cmd(*_PIDFILE_ARGS)
    def read_no_verify(self) -> None:
        with self._args_pidfile() as pidfile:
            print(pidfile.read())

    @ap.cmd(*_PIDFILE_ARGS)
    def lock(self) -> None:
        with self._args_pidfile() as pidfile:
            pidfile.acquire_lock()
            print(os.getpid())
            input()

    #

    _PIDFILE_PINNER_ARGS: ta.ClassVar[ta.Sequence[ap.Arg]] = [
        *_PIDFILE_ARGS,
        ap.arg('--timeout', type=float),
    ]

    def _pidfile_pinner_args(self) -> lang.Args:
        return self._pidfile_args().update(
            timeout=self._args.timeout,
        )

    def _args_pidfile_pinner(self) -> ta.ContextManager[int]:
        return self._pidfile_pinner_args()(PidfilePinner.default_impl()().pin_pidfile_owner)

    #

    @ap.cmd(*_PIDFILE_PINNER_ARGS)
    def read(self) -> None:
        with self._args_pidfile_pinner() as pid:
            print(pid)

    @ap.cmd(*_PIDFILE_PINNER_ARGS)
    def pin(self) -> None:
        with self._args_pidfile_pinner() as pid:
            print(pid)
            input()

    @ap.cmd(
        *_PIDFILE_PINNER_ARGS,
        ap.arg('signal'),
    )
    def kill(self) -> None:
        sig = parse_signal(self._args.signal)
        with self._args_pidfile_pinner() as pid:
            os.kill(pid, sig)


def _main() -> None:
    Cli().cli_run_and_exit()


if __name__ == '__main__':
    _main()



import ast
import typing as ta

from .codecs import make_object_lazy_loaded_codec
from .codecs import make_str_object_codec


##


def dumps(obj: ta.Any) -> str:
    return repr(obj)


def loads(s: str) -> ta.Any:
    return ast.literal_eval(s)


##


REPR_CODEC = make_str_object_codec('repr', dumps, loads)

# @omlish-manifest
_REPR_LAZY_CODEC = make_object_lazy_loaded_codec(__name__, 'REPR_CODEC', REPR_CODEC)



from .accessor import (  # noqa
    TypedValuesAccessor,
)

from .collection import (  # noqa
    DuplicateUniqueTypedValueError,

    TypedValues,

    collect,
    as_collection,
)

from .consumer import (  # noqa
    UnconsumedTypedValuesError,

    TypedValuesConsumer,

    consume,
)

from .generic import (  # noqa
    TypedValueGeneric,
)

from .holder import (  # noqa
    TypedValueHolder,
)

from .of_ import (  # noqa
    of,
)

from .reflect import (  # noqa
    reflect_typed_values_impls,
)

from .values import (  # noqa
    TypedValue,

    UniqueTypedValue,

    ScalarTypedValue,

    UniqueScalarTypedValue,
)


##


from ..lang.imports import _register_conditional_import  # noqa

_register_conditional_import('..marshal', '.marshal', __package__)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
Notes:
 - still racy as to if it's a different actual process as initial check, just with same pid, but due to 'identity' /
   semantic meaning of the named pidfile the processes are considered equivalent

Strategies:
 - linux
  - get pid of owner (lslocks or F_GETLK)
  - open pidfd to owner pid
  - re-check pid of owner
 - darwin
  - get pids of referrers (lsof)
  - read pid from file
  - ensure pid is in referrers
  - optionally loop
  - ? setup pid death watcher? still a race
"""
import abc
import contextlib
import os.path
import shutil
import sys
import time
import typing as ta

from ...diag.lslocks import LslocksCommand
from ...diag.lsof import LsofCommand
from ...lite.check import check
from ...lite.timeouts import Timeout
from ...lite.timeouts import TimeoutLike
from ...subprocesses.sync import subprocesses  # noqa
from .pidfile import Pidfile


##


class PidfilePinner(abc.ABC):
    def __init__(
            self,
            *,
            sleep_s: float = .1,
    ) -> None:
        super().__init__()

        self._sleep_s = sleep_s

    @classmethod
    @abc.abstractmethod
    def is_available(cls) -> bool:
        raise NotImplementedError

    class NoOwnerError(Exception):
        pass

    @abc.abstractmethod
    def _pin_pidfile_owner(self, pidfile: Pidfile, timeout: Timeout) -> ta.ContextManager[int]:
        raise NotImplementedError

    @contextlib.contextmanager
    def pin_pidfile_owner(
            self,
            path: str,
            *,
            timeout: TimeoutLike = None,
            inheritable: bool = False,  # Present to match Pidfile kwargs for convenience, but enforced to be False.
            **kwargs: ta.Any,
    ) -> ta.Iterator[int]:
        check.arg(not inheritable)

        timeout = Timeout.of(timeout)

        if not os.path.isfile(path):
            raise self.NoOwnerError

        with Pidfile(
                path,
                inheritable=False,
                **kwargs,
        ) as pf:
            try:
                with self._pin_pidfile_owner(pf, timeout) as pid:
                    yield pid

            except Pidfile.NotLockedError:
                raise self.NoOwnerError from None

    @classmethod
    def default_impl(cls) -> ta.Type['PidfilePinner']:
        for impl in [
            LslocksPidfdPidfilePinner,
            LsofPidfilePinner,
        ]:
            if impl.is_available():
                return impl
        return UnverifiedPidfilePinner


##


class UnverifiedPidfilePinner(PidfilePinner):
    @classmethod
    def is_available(cls) -> bool:
        return True

    @contextlib.contextmanager
    def _pin_pidfile_owner(self, pidfile: Pidfile, timeout: Timeout) -> ta.Iterator[int]:
        while (pid := pidfile.read()) is None:
            time.sleep(self._sleep_s)
            timeout()

        yield pid


##


class LsofPidfilePinner(PidfilePinner):
    """
    Fundamentally wrong, but still better than nothing. Simply reads the file contents and ensures a valid contained pid
    has the file open via `lsof`.
    """

    @classmethod
    def is_available(cls) -> bool:
        return shutil.which('lsof') is not None

    def _try_read_and_verify(self, pf: Pidfile, timeout: Timeout) -> ta.Optional[int]:
        if (initial_pid := pf.read()) is None:
            return None

        lsof_output = LsofCommand(
            # pid=initial_pid,
            file=os.path.abspath(pf.path),
        ).run(
            timeout=timeout,
        )

        lsof_pids: ta.Set[int] = set()
        for li in lsof_output:
            if li.pid is None:
                continue
            try:
                li_pid = int(li.pid)
            except ValueError:
                continue
            lsof_pids.add(li_pid)

        if initial_pid not in lsof_pids:
            return None

        if (reread_pid := pf.read()) is None or reread_pid != initial_pid:
            return None

        return reread_pid

    @contextlib.contextmanager
    def _pin_pidfile_owner(self, pidfile: Pidfile, timeout: Timeout) -> ta.Iterator[int]:
        while (pid := self._try_read_and_verify(pidfile, timeout)) is None:
            time.sleep(self._sleep_s)
            timeout()

        yield pid


##


class LslocksPidfdPidfilePinner(PidfilePinner):
    """
    Finds the locking pid via `lslocks`, opens a pidfd, then re-runs `lslocks` and rechecks the locking pid is the same.
    """

    @classmethod
    def is_available(cls) -> bool:
        return sys.platform == 'linux' and shutil.which('lslocks') is not None

    def _read_locking_pid(self, path: str, timeout: Timeout) -> int:
        lsl_output = LslocksCommand().run(timeout=timeout)

        lsl_pids = {
            li.pid
            for li in lsl_output
            if li.path == path
            and li.type == 'FLOCK'
        }
        if not lsl_pids:
            raise self.NoOwnerError
        if len(lsl_pids) != 1:
            raise RuntimeError(f'Multiple locks on file: {path}')

        [pid] = lsl_pids
        return pid

    class _Result(ta.NamedTuple):
        pid: int
        pidfd: int

    def _try_read_and_verify(
            self,
            pidfile: Pidfile,
            timeout: Timeout,
    ) -> ta.Optional[_Result]:
        path = os.path.abspath(pidfile.path)
        initial_pid = self._read_locking_pid(path, timeout)

        try:
            pidfd = os.open(f'/proc/{initial_pid}', os.O_RDONLY)
        except FileNotFoundError:
            raise self.NoOwnerError from None

        try:
            reread_pid = self._read_locking_pid(path, timeout)
            if reread_pid != initial_pid:
                os.close(pidfd)
                return None

            return self._Result(initial_pid, pidfd)

        except BaseException:
            os.close(pidfd)
            raise

    @contextlib.contextmanager
    def _pin_pidfile_owner(
            self,
            pidfile: Pidfile,
            timeout: Timeout,
    ) -> ta.Iterator[int]:
        while (res := self._try_read_and_verify(pidfile, timeout)) is None:
            time.sleep(self._sleep_s)
            timeout()

        try:
            yield res.pid
        finally:
            os.close(res.pidfd)


##


if __name__ == '__main__':
    def _main() -> None:
        argparse = __import__('argparse')
        parser = argparse.ArgumentParser()
        parser.add_argument('file')
        args = parser.parse_args()

        with PidfilePinner.default_impl()().pin_pidfile_owner(
                args.file,
                timeout=5.,
        ) as pid:
            print(pid)

    _main()



import typing as ta

from .... import lang
from .base import Backend
from .orjson import orjson_backend
from .std import std_backend
from .ujson import ujson_backend


##


@lang.cached_function
def default_backend() -> Backend:
    for fn in [
        orjson_backend,
        ujson_backend,
    ]:
        if (be := fn()) is not None:
            return be

    return std_backend()


##


def dump(obj: ta.Any, fp: ta.Any, **kwargs: ta.Any) -> None:
    return default_backend().dump(obj, fp, **kwargs)


def dumps(obj: ta.Any, **kwargs: ta.Any) -> str:
    return default_backend().dumps(obj, **kwargs)


def load(fp: ta.Any, **kwargs: ta.Any) -> ta.Any:
    return default_backend().load(fp, **kwargs)


def loads(s: str | bytes | bytearray, **kwargs: ta.Any) -> ta.Any:
    return default_backend().loads(s, **kwargs)


def dump_pretty(obj: ta.Any, fp: ta.Any, **kwargs: ta.Any) -> None:
    return default_backend().dump_pretty(obj, fp, **kwargs)


def dumps_pretty(obj: ta.Any, **kwargs: ta.Any) -> str:
    return default_backend().dumps_pretty(obj, **kwargs)


def dump_compact(obj: ta.Any, fp: ta.Any, **kwargs: ta.Any) -> None:
    return default_backend().dump_compact(obj, fp, **kwargs)


def dumps_compact(obj: ta.Any, **kwargs: ta.Any) -> str:
    return default_backend().dumps_compact(obj, **kwargs)



import dataclasses as dc
import inspect
import itertools
import types

from ..processing.base import Processor
from ..processing.priority import ProcessorPriority
from ..processing.registry import register_processor_type


##


def _dataclass_getstate(self):
    return [getattr(self, f.name) for f in dc.fields(self)]


def _dataclass_setstate(self, state):
    for field, value in zip(dc.fields(self), state):
        object.__setattr__(self, field.name, value)


def _get_slots(cls):
    match cls.__dict__.get('__slots__'):
        # A class which does not define __slots__ at all is equivalent to a class defining
        #   __slots__ = ('__dict__', '__weakref__')
        # `__dictoffset__` and `__weakrefoffset__` can tell us whether the base type has dict/weakref slots, in a way
        # that works correctly for both Python classes and C extension types. Extension types don't use `__slots__` for
        # slot creation
        case None:
            slots = []
            if getattr(cls, '__weakrefoffset__', -1) != 0:
                slots.append('__weakref__')
            if getattr(cls, '__dictoffset__', -1) != 0:
                slots.append('__dict__')
            yield from slots
        case str(slot):
            yield slot
        # Slots may be any iterable, but we cannot handle an iterator because it will already be (partially) consumed.
        case iterable if not hasattr(iterable, '__next__'):
            yield from iterable
        case _:
            raise TypeError(f"Slots of '{cls.__name__}' cannot be determined")


def _update_func_cell_for__class__(f, oldcls, newcls):
    # Returns True if we update a cell, else False.
    if f is None:
        # f will be None in the case of a property where not all of fget, fset, and fdel are used.  Nothing to do in
        # that case.
        return False
    try:
        idx = f.__code__.co_freevars.index('__class__')
    except ValueError:
        # This function doesn't reference __class__, so nothing to do.
        return False
    # Fix the cell to point to the new class, if it's already pointing at the old class.  I'm not convinced that the "is
    # oldcls" test is needed, but other than performance can't hurt.
    closure = f.__closure__[idx]
    if closure.cell_contents is oldcls:
        closure.cell_contents = newcls
        return True
    return False


def add_slots(
        cls: type,
        *,
        is_frozen: bool,
        weakref_slot: bool,
) -> type:
    # Need to create a new class, since we can't set __slots__ after a class has been created, and the @dataclass
    # decorator is called after the class is created.

    # Make sure __slots__ isn't already set.
    if '__slots__' in cls.__dict__:
        raise TypeError(f'{cls.__name__} already specifies __slots__')

    # Create a new dict for our new class.
    cls_dict = dict(cls.__dict__)
    field_names = tuple(f.name for f in dc.fields(cls))  # noqa

    # Make sure slots don't overlap with those in base classes.
    inherited_slots = set(itertools.chain.from_iterable(map(_get_slots, cls.__mro__[1:-1])))

    # The slots for our class.  Remove slots from our base classes.  Add '__weakref__' if weakref_slot was given, unless
    # it is already present.
    cls_dict['__slots__'] = tuple(
        itertools.filterfalse(
            inherited_slots.__contains__,
            itertools.chain(
                field_names,
                # gh-93521: '__weakref__' also needs to be filtered out if already present in inherited_slots
                ('__weakref__',) if weakref_slot else (),
            ),
        ),
    )

    for field_name in field_names:
        # Remove our attributes, if present. They'll still be available in _MARKER.
        cls_dict.pop(field_name, None)

    # Remove __dict__ itself.
    cls_dict.pop('__dict__', None)

    # Clear existing `__weakref__` descriptor, it belongs to a previous type:
    cls_dict.pop('__weakref__', None)

    qualname = getattr(cls, '__qualname__', None)
    newcls = type(cls)(cls.__name__, cls.__bases__, cls_dict)
    if qualname is not None:
        newcls.__qualname__ = qualname

    if is_frozen:
        # Need this for pickling frozen classes with slots.
        if '__getstate__' not in cls_dict:
            newcls.__getstate__ = _dataclass_getstate  # type: ignore
        if '__setstate__' not in cls_dict:
            newcls.__setstate__ = _dataclass_setstate  # type: ignore

    # Fix up any closures which reference __class__.  This is used to fix zero argument super so that it points to the
    # correct class (the newly created one, which we're returning) and not the original class.  We can break out of this
    # loop as soon as we make an update, since all closures for a class will share a given cell.
    for member in newcls.__dict__.values():
        # If this is a wrapped function, unwrap it.
        member = inspect.unwrap(member)
        if isinstance(member, types.FunctionType):
            if _update_func_cell_for__class__(member, cls, newcls):
                break

        elif isinstance(member, property):
            if (
                    _update_func_cell_for__class__(member.fget, cls, newcls) or
                    _update_func_cell_for__class__(member.fset, cls, newcls) or
                    _update_func_cell_for__class__(member.fdel, cls, newcls)
            ):
                break

    return newcls


##


@register_processor_type(priority=ProcessorPriority.SLOTS)
class SlotsProcessor(Processor):
    def check(self) -> None:
        if self._ctx.cs.weakref_slot and not self._ctx.cs.slots:
            raise TypeError('weakref_slot is True but slots is False')

    def process(self, cls: type) -> type:
        if not self._ctx.cs.slots:
            return cls

        return add_slots(
            cls,
            is_frozen=self._ctx.cs.frozen,
            weakref_slot=self._ctx.cs.weakref_slot,
        )



# type: ignore
# ruff: noqa
# flake8: noqa
# @omlish-generated
# Generated from Json5.g4 by ANTLR 4.13.2
from ....text.antlr._runtime._all import *
if "." in __name__:
    from .Json5Parser import Json5Parser
else:
    from Json5Parser import Json5Parser

# This class defines a complete generic visitor for a parse tree produced by Json5Parser.

class Json5Visitor(ParseTreeVisitor):

    # Visit a parse tree produced by Json5Parser#json5.
    def visitJson5(self, ctx:Json5Parser.Json5Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Json5Parser#obj.
    def visitObj(self, ctx:Json5Parser.ObjContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Json5Parser#pair.
    def visitPair(self, ctx:Json5Parser.PairContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Json5Parser#key.
    def visitKey(self, ctx:Json5Parser.KeyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Json5Parser#value.
    def visitValue(self, ctx:Json5Parser.ValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Json5Parser#arr.
    def visitArr(self, ctx:Json5Parser.ArrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Json5Parser#number.
    def visitNumber(self, ctx:Json5Parser.NumberContext):
        return self.visitChildren(ctx)



del Json5Parser


import dataclasses as dc
import typing as ta

from ..generation.base import Generator
from ..generation.base import Plan
from ..generation.base import PlanResult
from ..generation.ops import AddMethodOp
from ..generation.ops import Op
from ..generation.registry import register_generator_type
from ..generation.utils import build_attr_tuple_body_src_lines
from ..processing.base import ProcessingContext
from .fields import InstanceFields


##


NAME_OP_PAIRS = [
    ('__lt__', '<'),
    ('__le__', '<='),
    ('__gt__', '>'),
    ('__ge__', '>='),
]


##


@dc.dataclass(frozen=True)
class OrderPlan(Plan):
    fields: tuple[str, ...]


@register_generator_type(OrderPlan)
class OrderGenerator(Generator[OrderPlan]):
    def plan(self, ctx: ProcessingContext) -> PlanResult[OrderPlan] | None:
        if not ctx.cs.order:
            return None

        for name, _ in NAME_OP_PAIRS:
            if name in ctx.cls.__dict__:
                raise TypeError(
                    f'Cannot overwrite attribute {name} in class {ctx.cls.__name__}. '
                    f'Consider using functools.total_ordering',
                )

        return PlanResult(OrderPlan(
            tuple(f.name for f in ctx[InstanceFields] if f.compare),
        ))

    def generate(self, pl: OrderPlan) -> ta.Iterable[Op]:
        ops: list[AddMethodOp] = []

        for name, op in NAME_OP_PAIRS:
            ret_lines: list[str] = []
            if pl.fields:
                ret_lines.extend([
                    f'    return (',
                    *build_attr_tuple_body_src_lines(
                        'self',
                        *pl.fields,
                        prefix='        ',
                    ),
                    f'    ) {op} (',
                    *build_attr_tuple_body_src_lines(
                        'other',
                        *pl.fields,
                        prefix='        ',
                    ),
                    f'    )',
                ])
            else:
                ret_lines.append(
                    f'    return {"True" if "=" in op else "False"}',
                )

            ops.append(AddMethodOp(
                name,
                '\n'.join([
                    f'def {name}(self, other):',
                    f'    if other.__class__ is not self.__class__:',
                    f'        return NotImplemented',
                    *ret_lines,
                ]),
            ))

        return ops



import functools
import typing as ta

from .. import lang


if ta.TYPE_CHECKING:
    import asyncio

    import sniffio
    import trio_asyncio

else:
    asyncio = lang.proxy_import('asyncio')

    sniffio = lang.proxy_import('sniffio')
    trio_asyncio = lang.proxy_import('trio_asyncio')


##


def check_trio_asyncio() -> None:
    if trio_asyncio.current_loop.get() is None:
        raise RuntimeError('trio_asyncio loop not running')


def with_trio_asyncio_loop(*, wait=False, strict=False):
    def outer(fn):
        @functools.wraps(fn)
        async def inner(*args, **kwargs):
            if trio_asyncio.current_loop.get() is not None:
                await fn(*args, **kwargs)
                return

            if sniffio.current_async_library() != 'trio':
                if strict:
                    raise RuntimeError('trio loop not running')
                await fn(*args, **kwargs)
                return

            loop: asyncio.BaseEventLoop
            async with trio_asyncio.open_loop() as loop:
                try:
                    await fn(*args, **kwargs)
                finally:
                    if wait:
                        # FIXME: lol
                        while asyncio.all_tasks(loop):
                            await asyncio.sleep(.1)

        return inner

    return outer



"""
TODO:
 - len
 - required character classes
  - lowercase
  - uppercase
  - digits
  - symbols
 - move to omlish/secrets
 - argparse, CliCmd
"""
import argparse
import random
import secrets
import string
import typing as ta


##


CHAR_CLASSES: ta.Mapping[str, str] = {
    'lower': string.ascii_lowercase,
    'upper': string.ascii_uppercase,
    'digit': string.digits,
    'special': string.punctuation,
}


ALL_CHAR_CLASSES = tuple(CHAR_CLASSES.values())
DEFAULT_LENGTH = 16


def generate_password(
        char_classes: ta.Sequence[str] = ALL_CHAR_CLASSES,
        length: int = DEFAULT_LENGTH,
        *,
        rand: random.Random | None = None,
) -> str:
    if rand is None:
        rand = secrets.SystemRandom()
    l: list[str] = []
    for cc in char_classes:
        l.append(rand.choice(cc))
    cs = ''.join(char_classes)
    if not cs:
        raise ValueError(cs)
    while len(l) < length:
        l.append(rand.choice(cs))
    rand.shuffle(l)
    return ''.join(l)


def _main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument('length', type=int, nargs='?', default=DEFAULT_LENGTH)
    for cc in CHAR_CLASSES:
        parser.add_argument(f'-{cc[0]}', f'--{cc}', action='store_true')
    args = parser.parse_args()

    cs = {
        cc
        for cc in CHAR_CLASSES
        if getattr(args, cc) is not None
    }
    if cs:
        ccs = tuple(CHAR_CLASSES[cc] for cc in cs)
    else:
        ccs = ALL_CHAR_CLASSES

    pw = generate_password(
        ccs,
        args.length,
    )
    print(pw)


# @omlish-manifest
_CLI_MODULE = {'$omdev.cli.types.CliModule': {
    'cmd_name': 'pwgen',
    'mod_name': __name__,
}}


if __name__ == '__main__':
    _main()



import functools
import json
import typing as ta


##


JSON_PRETTY_INDENT = 2

JSON_PRETTY_KWARGS: ta.Mapping[str, ta.Any] = dict(
    indent=JSON_PRETTY_INDENT,
)

json_dump_pretty: ta.Callable[..., None] = functools.partial(json.dump, **JSON_PRETTY_KWARGS)
json_dumps_pretty: ta.Callable[..., str] = functools.partial(json.dumps, **JSON_PRETTY_KWARGS)


##


JSON_COMPACT_SEPARATORS = (',', ':')

JSON_COMPACT_KWARGS: ta.Mapping[str, ta.Any] = dict(
    indent=None,
    separators=JSON_COMPACT_SEPARATORS,
)

json_dump_compact: ta.Callable[..., None] = functools.partial(json.dump, **JSON_COMPACT_KWARGS)
json_dumps_compact: ta.Callable[..., str] = functools.partial(json.dumps, **JSON_COMPACT_KWARGS)



"""
TODO:
 - Config? dedupe defaults with Daemon
 - ExitStacked? hold ref to Spawner, which holds refs to thread/proc - which will likely outlive it, but still
"""
import contextlib
import functools
import logging
import threading
import typing as ta

from .. import check
from .. import lang
from ..os.pidfiles.manager import open_inheritable_pidfile
from ..os.pidfiles.pidfile import Pidfile
from .reparent import reparent_process
from .spawning import InProcessSpawner
from .spawning import Spawn
from .spawning import Spawner
from .spawning import Spawning
from .spawning import spawner_for
from .targets import Target
from .targets import target_runner_for


log = logging.getLogger(__name__)


##


class Launcher:
    def __init__(
            self,
            *,
            target: Target,
            spawning: Spawning,

            pid_file: str | None = None,
            reparent_process: bool = False,  # noqa
            launched_timeout_s: float = 5.,
    ) -> None:
        super().__init__()

        self._target = target
        self._spawning = spawning

        self._pid_file = pid_file
        self._reparent_process = reparent_process
        self._launched_timeout_s = launched_timeout_s

    def _inner_launch(
            self,
            *,
            pidfile_manager: ta.ContextManager | None,
            callback: ta.Callable[[], None] | None = None,
    ) -> None:
        callback_called = False

        try:
            if self._reparent_process:
                log.info('Reparenting')
                reparent_process()

            with contextlib.ExitStack() as es:
                pidfile: Pidfile | None = None  # noqa
                if pidfile_manager is not None:
                    pidfile = check.isinstance(es.enter_context(pidfile_manager), Pidfile)
                    pidfile.write()

                if callback is not None:
                    callback_called = True
                    callback()

                runner = target_runner_for(self._target)
                runner.run()

        finally:
            if callback is not None and not callback_called:
                callback()

    def launch(self) -> bool:
        with contextlib.ExitStack() as es:
            spawner: Spawner = es.enter_context(spawner_for(self._spawning))

            #

            inherit_fds: set[int] = set()
            launched_event: threading.Event | None = None

            pidfile: Pidfile | None = None  # noqa
            pidfile_manager: ta.ContextManager | None = None

            if (pid_file := self._pid_file) is not None:
                if not isinstance(spawner, InProcessSpawner):
                    pidfile = es.enter_context(open_inheritable_pidfile(pid_file))
                    pidfile_manager = lang.NopContextManager(pidfile)

                else:
                    check.state(not self._reparent_process)
                    pidfile = es.enter_context(Pidfile(pid_file))
                    pidfile_manager = pidfile.dup()
                    launched_event = threading.Event()

                if not pidfile.try_acquire_lock():
                    return False

                inherit_fds.add(check.isinstance(pidfile.fileno(), int))

            #

            spawner.spawn(Spawn(
                functools.partial(
                    self._inner_launch,
                    pidfile_manager=pidfile_manager,
                    callback=launched_event.set if launched_event is not None else None,
                ),
                target=self._target,
                inherit_fds=inherit_fds,
            ))

            if launched_event is not None:
                check.state(launched_event.wait(timeout=self._launched_timeout_s))

            return True



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import dataclasses as dc
import os
import typing as ta

from ..lite.check import check
from ..lite.timeouts import Timeout
from ..subprocesses.run import SubprocessRun
from ..subprocesses.run import SubprocessRunnable
from ..subprocesses.run import SubprocessRunOutput
from ..subprocesses.sync import subprocesses


##


@dc.dataclass(frozen=True)
class PsItem:
    pid: int
    ppid: int
    cmd: str


@dc.dataclass(frozen=True)
class PsCommand(SubprocessRunnable):
    pid: ta.Optional[int] = None

    def make_run(self) -> SubprocessRun:
        return SubprocessRun.of(
            'ps',
            '-o', 'pid=,ppid=,command=',
            *([str(int(self.pid))] if self.pid is not None else []),

            check=True,
            stdout='pipe',
            stderr='devnull',
        )

    def handle_run_output(self, output: SubprocessRunOutput) -> PsItem:
        opid, ppid, cmd = check.not_none(output.stdout).decode().split(maxsplit=2)
        return PsItem(
            int(opid),
            int(ppid),
            cmd.strip(),
        )


##


def get_ps_item(pid: int, timeout: ta.Optional[Timeout] = None) -> PsItem:
    return PsCommand(pid).run(subprocesses, timeout=timeout)


def get_ps_lineage(pid: int, timeout: ta.Optional[Timeout] = None) -> ta.List[PsItem]:
    timeout = Timeout.of(timeout)
    ret: list[PsItem] = []
    while True:
        cur = get_ps_item(pid, timeout)
        if cur.ppid < 1:
            break
        ret.append(cur)
        pid = cur.ppid
    return ret


##


if __name__ == '__main__':
    def _main() -> None:
        print(get_ps_lineage(os.getpid()))

    _main()



# type: ignore
# ruff: noqa
# flake8: noqa
#
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
#

#
#  This is an InputStream that is loaded from a file all at once
#  when you construct the object.
#

import codecs
from .InputStream import InputStream


class FileStream(InputStream):
    __slots__ = 'fileName'

    def __init__(self, fileName:str, encoding:str='ascii', errors:str='strict'):
        super().__init__(self.readDataFrom(fileName, encoding, errors))
        self.fileName = fileName

    def readDataFrom(self, fileName:str, encoding:str, errors:str='strict'):
        # read binary to avoid line ending conversion
        with open(fileName, 'rb') as file:
            bytes = file.read()
            return codecs.decode(bytes, encoding, errors)



import typing as ta

from .. import dataclasses as dc
from .. import lang
from .impl.origins import HasOriginsImpl


T = ta.TypeVar('T')


##


@dc.dataclass(frozen=True, eq=False)
class ConstFn(HasOriginsImpl, lang.Final, ta.Generic[T]):
    """An origin tracking provider function for a constant value. Equivalent to `lambda: v` but transparent."""

    v: T

    def __call__(self) -> T:
        return self.v


##


@dc.dataclass(frozen=True)
@dc.extra_class_params(terse_repr=True)
class Id:
    """A utility dataclass intended to be used as a key tag for disambiguation."""

    v: int



import inspect
import typing as ta

from .. import dataclasses as dc
from .. import lang
from .. import reflect as rfl
from .types import Tag


T = ta.TypeVar('T')


##


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class Key(lang.Final, ta.Generic[T]):
    ty: rfl.Type = dc.xfield(coerce=rfl.type_)

    tag: ta.Any = dc.xfield(
        default=None,
        kw_only=True,
        validate=lambda o: not isinstance(o, Tag),
        repr_fn=dc.opt_repr,
    )


def as_key(o: ta.Any) -> Key:
    if o is None or o is inspect.Parameter.empty:
        raise TypeError(o)
    if isinstance(o, Key):
        return o
    return Key(rfl.type_(o))



"""
A background thread that opens a unix socket server accepting connections to an in-proc python repl (from which one can
inspect module globals, live thread stacks, and other such things).
"""
from .server import ReplServer  # noqa



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import io
import sys
import traceback
import types
import typing as ta


##


class LoggingCaller(ta.NamedTuple):
    filename: str
    lineno: int
    func: str
    sinfo: ta.Optional[str]

    @classmethod
    def find_frame(cls, ofs: int = 0) -> types.FrameType:
        f: ta.Any = sys._getframe(2 + ofs)  # noqa
        while hasattr(f, 'f_code'):
            if f.f_code.co_filename != __file__:
                return f
            f = f.f_back
        raise RuntimeError

    @classmethod
    def find(cls, stack_info: bool = False) -> 'LoggingCaller':
        f = cls.find_frame(1)
        # TODO: ('(unknown file)', 0, '(unknown function)', None) ?

        sinfo = None
        if stack_info:
            sio = io.StringIO()
            sio.write('Stack (most recent call last):\n')
            traceback.print_stack(f, file=sio)
            sinfo = sio.getvalue()
            sio.close()
            if sinfo[-1] == '\n':
                sinfo = sinfo[:-1]

        return cls(
            f.f_code.co_filename,
            f.f_lineno,
            f.f_code.co_name,
            sinfo,
        )



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
TODO:
 - cascading?
 - caching?
 - purging?
 - wishlist: piecewise config parts, unmarshal, registered via manifests
"""
import abc
import os.path
import typing as ta

from ..lite.check import check
from ..os.mangle import mangle_path
from .formats import DEFAULT_CONFIG_LOADERS
from .formats import ConfigLoader


ShadowConfig = ta.Mapping[str, ta.Any]  # ta.TypeAlias


##


class ShadowConfigs(abc.ABC):
    @abc.abstractmethod
    def get_shadow_config(self, path: str) -> ta.Optional[ShadowConfig]:
        raise NotImplementedError


class FileShadowConfigs(ShadowConfigs, abc.ABC):
    @abc.abstractmethod
    def get_shadow_config_file_path(self, path: str) -> str:
        raise NotImplementedError


##


class NopShadowConfigs(ShadowConfigs):
    def get_shadow_config(self, path: str) -> ta.Optional[ShadowConfig]:
        return None


##


class MangledFilesShadowConfigs(FileShadowConfigs):
    def __init__(
            self,
            path: str,
            *,
            create: bool = False,
            ext: str = 'yml',
            loader: ta.Optional[ConfigLoader] = None,
    ) -> None:  # noqa
        super().__init__()

        check.arg(not ext.startswith('.'))
        check.non_empty_str(ext)

        if loader is None:
            for dl in DEFAULT_CONFIG_LOADERS:
                if dl.match_file(f'.{ext}'):
                    loader = dl
                    break
            else:
                raise KeyError(ext)

        self._path = path
        self._create = create
        self._ext = ext
        self._loader = loader

    def get_shadow_config_file_path(self, path: str) -> str:
        mangled = '.'.join([mangle_path(os.path.abspath(path)), self._ext])
        return os.path.abspath(os.path.join(self._path, mangled))

    def get_shadow_config(self, path: str) -> ta.Optional[ShadowConfig]:
        file_path = self.get_shadow_config_file_path(path)

        if self._create:
            os.makedirs(self._path, exist_ok=True)

        try:
            with open(file_path) as f:
                content = f.read()
        except FileNotFoundError:
            return None

        return self._loader.load_str(content).as_map()



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
See:
 - https://nginx.org/en/docs/dev/development_guide.html
 - https://nginx.org/en/docs/dev/development_guide.html#config_directives
 - https://nginx.org/en/docs/example.html
"""
import collections.abc
import dataclasses as dc
import typing as ta

from ..lite.check import check
from ..text.indent import IndentWriter


##


@dc.dataclass()
class NginxConfigItems:
    lst: ta.List['NginxConfigItem']

    @classmethod
    def of(cls, obj: ta.Any) -> 'NginxConfigItems':
        if isinstance(obj, NginxConfigItems):
            return obj
        return cls([NginxConfigItem.of(e) for e in check.isinstance(obj, list)])


@dc.dataclass()
class NginxConfigItem:
    name: str
    args: ta.Optional[ta.List[str]] = None
    block: ta.Optional[NginxConfigItems] = None

    @classmethod
    def of(cls, obj: ta.Any) -> 'NginxConfigItem':
        if isinstance(obj, NginxConfigItem):
            return obj
        args = check.isinstance(check.not_isinstance(obj, str), collections.abc.Sequence)
        name, args = check.isinstance(args[0], str), args[1:]
        if args and not isinstance(args[-1], str):
            block, args = NginxConfigItems.of(args[-1]), args[:-1]
        else:
            block = None
        return NginxConfigItem(name, [check.isinstance(e, str) for e in args], block=block)


def render_nginx_config(wr: IndentWriter, obj: ta.Any) -> None:
    if isinstance(obj, NginxConfigItem):
        wr.write(obj.name)
        for e in obj.args or ():
            wr.write(' ')
            wr.write(e)
        if obj.block:
            wr.write(' {\n')
            with wr.indent():
                render_nginx_config(wr, obj.block)
            wr.write('}\n')
        else:
            wr.write(';\n')

    elif isinstance(obj, NginxConfigItems):
        for e2 in obj.lst:
            render_nginx_config(wr, e2)

    else:
        raise TypeError(obj)


def render_nginx_config_str(obj: ta.Any) -> str:
    iw = IndentWriter()
    render_nginx_config(iw, obj)
    return iw.getvalue()



import argparse
import dataclasses as dc
import sys
import typing as ta

from ...formats import json
from .ast import Node
from .errors import ArityError
from .errors import JmespathTypeError
from .errors import JmespathValueError
from .errors import ParseError
from .errors import UnknownFunctionError
from .parser import compile  # noqa
from .parser import search
from .visitor import node_type


##


def _ast_to_json(o: ta.Any) -> ta.Any:
    if isinstance(o, json.SCALAR_TYPES):
        return o
    elif isinstance(o, Node):
        return {node_type(o): {f.name: _ast_to_json(getattr(o, f.name)) for f in dc.fields(o)}}
    elif isinstance(o, (list, tuple)):
        return [_ast_to_json(e) for e in o]
    else:
        raise TypeError(o)


def _main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument('expression')
    parser.add_argument('-f', '--filename')
    parser.add_argument('-c', '--compact', action='store_true')
    parser.add_argument('--ast', action='store_true')
    args = parser.parse_args()

    if args.compact:
        json_dumps = json.dumps_compact
    else:
        json_dumps = json.dumps_pretty

    expression = args.expression
    if args.ast:
        expression = compile(args.expression)
        print(json_dumps(_ast_to_json(expression.parsed)))
        return 0

    if args.filename:
        with open(args.filename) as f:
            data = json.load(f)
    else:
        data = sys.stdin.read()
        data = json.loads(data)

    try:
        print(json_dumps(search(expression, data)))
        return 0

    except ArityError as e:
        print(f'invalid-arity: {e}', file=sys.stderr)
        return 1

    except JmespathTypeError as e:
        print(f'invalid-type: {e}', file=sys.stderr)
        return 1

    except JmespathValueError as e:
        print(f'invalid-value: {e}', file=sys.stderr)
        return 1

    except UnknownFunctionError as e:
        print(f'unknown-function: {e}', file=sys.stderr)
        return 1

    except ParseError as e:
        print(f'syntax-error: {e}', file=sys.stderr)
        return 1


if __name__ == '__main__':
    sys.exit(_main())



import typing as _ta

from .asyncio import AsyncioAsyncsBackend  # noqa
from .base import AsyncsBackend  # noqa
from .trio import TrioAsyncsBackend  # noqa
from .trio_asyncio import TrioAsyncioAsyncsBackend  # noqa


##


ASYNC_BACKENDS: _ta.Collection[type[AsyncsBackend]] = [
    AsyncioAsyncsBackend,
    TrioAsyncioAsyncsBackend,
    TrioAsyncsBackend,
]



"""
Analogous to scala's partial functions ( https://www.scala-lang.org/api/current/scala/PartialFunction.html ).

TODO:
 - unify MatchFnClass with dispatch.method?
  - __call__ = mfs.method(); @__call__.register(lambda: ...) def _call_... ?
  - not really the same thing, dispatch is unordered this is necessarily ordered
"""
import abc
import dataclasses as dc
import typing as ta

from .. import lang


T = ta.TypeVar('T')
P = ta.ParamSpec('P')


##


class MatchGuardError(Exception):
    pass


class MatchFn(abc.ABC, ta.Generic[P, T]):
    @abc.abstractmethod
    def guard(self, *args: P.args, **kwargs: P.kwargs) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def fn(self, *args: P.args, **kwargs: P.kwargs) -> T:
        raise NotImplementedError

    def __get__(self, instance, owner=None):
        return self

    @ta.final
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        if not self.guard(*args, **kwargs):
            raise MatchGuardError(*args, **kwargs)
        return self.fn(*args, **kwargs)


##


@dc.dataclass(frozen=True)
class SimpleMatchFn(MatchFn[P, T]):
    _guard: ta.Callable[P, bool]
    _fn: ta.Callable[P, T]

    def guard(self, *args: P.args, **kwargs: P.kwargs) -> bool:
        return self._guard(*args, **kwargs)

    def fn(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self._fn(*args, **kwargs)

    def __get__(self, instance, owner=None):
        return self.__class__(
            self._guard.__get__(instance, owner),  # noqa
            self._fn.__get__(instance, owner),  # noqa
        )


@ta.overload
def simple(guard: ta.Callable[..., bool], fn: ta.Callable[P, T]) -> SimpleMatchFn[P, T]:
    ...


@ta.overload
def simple(guard: ta.Callable[..., bool]) -> ta.Callable[[ta.Callable[P, T]], SimpleMatchFn[P, T]]:
    ...


def simple(guard, fn=None):
    def inner(fn):  # noqa
        return SimpleMatchFn(guard, fn)
    if fn is not None:
        return inner(fn)
    else:
        return inner


##


class AmbiguousMatchesError(Exception):
    pass


@dc.dataclass(frozen=True)
class MultiMatchFn(MatchFn[P, T]):
    children: ta.Sequence[MatchFn[P, T]]
    strict: bool = False

    def match(self, *args: P.args, **kwargs: P.kwargs) -> MatchFn[P, T] | None:
        matches = []
        for cur in self.children:
            if cur.guard(*args, **kwargs):
                if self.strict:
                    matches.append(cur)
                else:
                    return cur
        if not matches:
            return None
        elif len(matches) > 1:
            raise AmbiguousMatchesError
        else:
            return matches[0]

    def guard(self, *args: P.args, **kwargs: P.kwargs) -> bool:
        return self.match(*args, **kwargs) is not None

    def fn(self, *args: P.args, **kwargs: P.kwargs) -> T:
        if (m := self.match(*args, **kwargs)) is None:
            raise MatchGuardError(*args, **kwargs)
        return m.fn(*args, **kwargs)

    def __get__(self, instance, owner=None):
        return self.__class__(
            [c.__get__(instance, owner) for c in self.children],
            strict=self.strict,
        )


def multi(*children: MatchFn[P, T], strict: bool = False) -> MultiMatchFn:  # MultiMatchFn[P[0], T[-1]]
    return MultiMatchFn(children, strict=strict)  # noqa


##


class CachedMatchFn(MatchFn[P, T]):
    @staticmethod
    def _default_key(*args, **kwargs):
        return (args, tuple(sorted(kwargs.items(), key=lambda t: t[0])))

    def __init__(
            self,
            f: MatchFn[P, T],
            *,
            key: ta.Callable[P, ta.Any] = _default_key,
            lock: lang.DefaultLockable = None,
    ) -> None:
        super().__init__()

        self._f = f
        self._key = key
        self._lock = lock
        self._lock_impl = lang.default_lock(lock)()
        self._dct: dict[ta.Any, lang.Maybe[ta.Any]] = {}

    def guard(self, *args: P.args, **kwargs: P.kwargs) -> bool:
        with self._lock_impl:
            k = self._key(*args, **kwargs)
            try:
                e = self._dct[k]
            except KeyError:
                if self._f.guard(*args, **kwargs):
                    return True
                else:
                    self._dct[k] = lang.empty()
                    return False
            else:
                return e.present

    def fn(self, *args: P.args, **kwargs: P.kwargs) -> T:
        with self._lock_impl:
            k = self._key(*args, **kwargs)
            try:
                e = self._dct[k]
            except KeyError:
                try:
                    ret = self._f(*args, **kwargs)
                except MatchGuardError:
                    self._dct[k] = lang.empty()
                    raise
                else:
                    self._dct[k] = lang.just(ret)
                    return ret
            else:
                if e.present:
                    return e.must()
                else:
                    raise MatchGuardError(*args, **kwargs)

    def __get__(self, instance, owner=None):
        return self.__class__(self._f.__get__(instance, owner), key=self._key)  # noqa


cached = CachedMatchFn


##


class MatchFnClass(MatchFn[P, T]):
    _cls_match_fn: ta.ClassVar[MultiMatchFn]

    __match_fn: MatchFn[P, T] | None = None

    @property
    def _match_fn(self) -> MatchFn[P, T]:
        if (mf := self.__match_fn) is None:
            mf = self.__match_fn = self._cls_match_fn.__get__(self)
        return mf

    def __init_subclass__(cls, strict: bool = False, **kwargs: ta.Any) -> None:
        super().__init_subclass__()

        if '_cls_match_fn' in cls.__dict__:
            raise AttributeError('_cls_match_fn')

        d = {}
        for c in cls.__mro__:
            for a, o in c.__dict__.items():
                if isinstance(o, MatchFn) and a not in d:
                    d[a] = o
        cls._cls_match_fn = MultiMatchFn(list(d.values()), strict=strict)

    def guard(self, *args: P.args, **kwargs: P.kwargs) -> bool:
        return self._match_fn.guard(*args, **kwargs)

    def fn(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self._match_fn.fn(*args, **kwargs)



import itertools
import operator


##


def truncate_utf8_safe(string: str, length: int) -> str:
    return string[:length] \
        .encode('utf-8', 'ignore') \
        .decode('utf-8', 'ignore')


def find_utf8_truncation_point(buf: bytes | bytearray, length: int) -> int:
    if len(buf) < length:
        raise ValueError(buf)
    cs = itertools.accumulate(
        (len(c.encode('utf-8')) for c in buf.decode('utf-8')),
        operator.add,
        initial=0,
    )
    return next(i for i, o in enumerate(cs) if o >= length)


def trim_initial_spaces(string: str) -> str:
    return string.lstrip(' ')


def is_ascii(string: str) -> bool:
    return all(ord(c) < 128 for c in string)



from .classes.decorator import (  # noqa
    dataclass,
)

from .classes.make import (  # noqa
    make_dataclass,
)

from .classes.metadata import (  # noqa
    append_class_metadata,
    extra_class_params,
    init,
    metadata,
    validate,
)

from .fields.metadata import (  # noqa
    extra_field_params,
    set_field_metadata,
    update_extra_field_params,
    with_extra_field_params,
)

from .fields.constructor import (  # noqa
    field,
)



from .types import NumericReplies
from .types import NumericReply


##


_REGISTERED_NUMERIC_REPLIES: list[NumericReply] = []


def _register_numeric_reply(
        name: str,
        num: int,
        *formats: str,
) -> NumericReply:
    nr = NumericReply(
        name,
        num,
        formats,
    )
    _REGISTERED_NUMERIC_REPLIES.append(nr)
    return nr


##


RPL_WELCOME = _register_numeric_reply(
    'RPL_WELCOME',
    1,
    '<client> :Welcome to the <networkname> Network, <nick>[!<user>@<host>]',
)

RPL_YOURHOST = _register_numeric_reply(
    'RPL_YOURHOST',
    2,
    '<client> :Your host is <servername>, running version <version>',
)

RPL_CREATED = _register_numeric_reply(
    'RPL_CREATED',
    3,
    '<client> :This server was created <datetime>',
)

RPL_MYINFO = _register_numeric_reply(
    'RPL_MYINFO',
    4,
    '<client> <servername> <version> <available user modes> <available channel modes> [<channel modes with a parameter>]',  # noqa
    # noqa
)

RPL_ISUPPORT = _register_numeric_reply(
    'RPL_ISUPPORT',
    5,
    '<client> <1-13 tokens> :are supported by this server',
)

RPL_BOUNCE = _register_numeric_reply(
    'RPL_BOUNCE',
    10,
    '<client> <hostname> <port> :<info>',
)

RPL_STATSLINKINFO = _register_numeric_reply(
    'RPL_STATSLINKINFO',
    211,
    '<linkname> <sendq> <sent messages> <sent Kbytes> <received messages> <received Kbytes> <time open>',
)

RPL_STATSCOMMANDS = _register_numeric_reply(
    'RPL_STATSCOMMANDS',
    212,
    '<client> <command> <count> [<byte count> <remote count>]',
)

RPL_STATSCLINE = _register_numeric_reply('RPL_STATSCLINE', 213)

RPL_STATSNLINE = _register_numeric_reply('RPL_STATSNLINE', 214)

RPL_STATSILINE = _register_numeric_reply('RPL_STATSILINE', 215)

RPL_STATSKLINE = _register_numeric_reply('RPL_STATSKLINE', 216)

RPL_STATSQLINE = _register_numeric_reply('RPL_STATSQLINE', 217)

RPL_STATSYLINE = _register_numeric_reply('RPL_STATSYLINE', 218)

RPL_ENDOFSTATS = _register_numeric_reply(
    'RPL_ENDOFSTATS',
    219,
    '<client> <stats letter> :End of /STATS report',
)

RPL_UMODEIS = _register_numeric_reply(
    'RPL_UMODEIS',
    221,
    '<client> <user modes>',
)

RPL_STATSVLINE = _register_numeric_reply('RPL_STATSVLINE', 240)

RPL_STATSLLINE = _register_numeric_reply('RPL_STATSLLINE', 241)

RPL_STATSUPTIME = _register_numeric_reply(
    'RPL_STATSUPTIME',
    242,
    '<client> :Server Up <days> days <hours>:<minutes>:<seconds>',
)

RPL_STATSOLINE = _register_numeric_reply(
    'RPL_STATSOLINE',
    243,
    'O <hostmask> * <name>',
)

RPL_STATSHLINE = _register_numeric_reply('RPL_STATSHLINE', 244)

RPL_STATSPING = _register_numeric_reply('RPL_STATSPING', 246)

RPL_STATSBLINE = _register_numeric_reply('RPL_STATSBLINE', 247)

RPL_STATSDLINE = _register_numeric_reply('RPL_STATSDLINE', 250)

RPL_LUSERCLIENT = _register_numeric_reply(
    'RPL_LUSERCLIENT',
    251,
    '<client> :There are <u> users and <i> invisible on <s> servers',
)

RPL_LUSEROP = _register_numeric_reply(
    'RPL_LUSEROP',
    252,
    '<client> <ops> :operator(s) online',
)

RPL_LUSERUNKNOWN = _register_numeric_reply(
    'RPL_LUSERUNKNOWN',
    253,
    '<client> <connections> :unknown connection(s)',
)

RPL_LUSERCHANNELS = _register_numeric_reply(
    'RPL_LUSERCHANNELS',
    254,
    '<client> <channels> :channels formed',
)

RPL_LUSERME = _register_numeric_reply(
    'RPL_LUSERME',
    255,
    '<client> :I have <c> clients and <s> servers',
)

RPL_ADMINME = _register_numeric_reply(
    'RPL_ADMINME',
    256,
    '<client> [<server>] :Administrative info',
)

RPL_ADMINLOC1 = _register_numeric_reply(
    'RPL_ADMINLOC1',
    257,
    '<client> :<info>',
)

RPL_ADMINLOC2 = _register_numeric_reply(
    'RPL_ADMINLOC2',
    258,
    '<client> :<info>',
)

RPL_ADMINEMAIL = _register_numeric_reply(
    'RPL_ADMINEMAIL',
    259,
    '<client> :<info>',
)

RPL_TRYAGAIN = _register_numeric_reply(
    'RPL_TRYAGAIN',
    263,
    '<client> <command> :Please wait a while and try again.',
)

RPL_LOCALUSERS = _register_numeric_reply(
    'RPL_LOCALUSERS',
    265,
    '<client> [<u> <m>] :Current local users <u>, max <m>',
)

RPL_GLOBALUSERS = _register_numeric_reply(
    'RPL_GLOBALUSERS',
    266,
    '<client> [<u> <m>] :Current global users <u>, max <m>',
)

RPL_WHOISCERTFP = _register_numeric_reply(
    'RPL_WHOISCERTFP',
    276,
    '<client> <nick> :has client certificate fingerprint <fingerprint>',
)

RPL_NONE = _register_numeric_reply(
    'RPL_NONE',
    300,
)

RPL_AWAY = _register_numeric_reply(
    'RPL_AWAY',
    301,
    '<client> <nick> :<message>',
)

RPL_USERHOST = _register_numeric_reply(
    'RPL_USERHOST',
    302,
    '<client> :[<reply>{ <reply>}]',
)

RPL_UNAWAY = _register_numeric_reply(
    'RPL_UNAWAY',
    305,
    '<client> :You are no longer marked as being away',
)

RPL_NOWAWAY = _register_numeric_reply(
    'RPL_NOWAWAY',
    306,
    '<client> :You have been marked as being away',
)

RPL_WHOISREGNICK = _register_numeric_reply(
    'RPL_WHOISREGNICK',
    307,
    '<client> <nick> :has identified for this nick',
)

RPL_WHOISUSER = _register_numeric_reply(
    'RPL_WHOISUSER',
    311,
    '<client> <nick> <username> <host> * :<realname>',
)

RPL_WHOISSERVER = _register_numeric_reply(
    'RPL_WHOISSERVER',
    312,
    '<client> <nick> <server> :<server info>',
)

RPL_WHOISOPERATOR = _register_numeric_reply(
    'RPL_WHOISOPERATOR',
    313,
    '<client> <nick> :is an IRC operator',
)

RPL_WHOWASUSER = _register_numeric_reply(
    'RPL_WHOWASUSER',
    314,
    '<client> <nick> <username> <host> * :<realname>',
)

RPL_ENDOFWHO = _register_numeric_reply(
    'RPL_ENDOFWHO',
    315,
    '<client> <mask> :End of WHO list',
)

RPL_WHOISIDLE = _register_numeric_reply(
    'RPL_WHOISIDLE',
    317,
    '<client> <nick> <secs> <signon> :seconds idle, signon time',
)

RPL_ENDOFWHOIS = _register_numeric_reply(
    'RPL_ENDOFWHOIS',
    318,
    '<client> <nick> :End of /WHOIS list',
)

RPL_WHOISCHANNELS = _register_numeric_reply(
    'RPL_WHOISCHANNELS',
    319,
    '<client> <nick> :[prefix]<channel>{ [prefix]<channel>}',
)

RPL_WHOISSPECIAL = _register_numeric_reply(
    'RPL_WHOISSPECIAL',
    320,
    '<client> <nick> :blah blah blah',
)

RPL_LISTSTART = _register_numeric_reply(
    'RPL_LISTSTART',
    321,
    '<client> Channel :Users  Name',
)

RPL_LIST = _register_numeric_reply(
    'RPL_LIST',
    322,
    '<client> <channel> <client count> :<topic>',
)

RPL_LISTEND = _register_numeric_reply(
    'RPL_LISTEND',
    323,
    '<client> :End of /LIST',
)

RPL_CHANNELMODEIS = _register_numeric_reply(
    'RPL_CHANNELMODEIS',
    324,
    '<client> <channel> <modestring> <mode arguments>...',
)

RPL_CREATIONTIME = _register_numeric_reply(
    'RPL_CREATIONTIME',
    329,
    '<client> <channel> <creationtime>',
)

RPL_WHOISACCOUNT = _register_numeric_reply(
    'RPL_WHOISACCOUNT',
    330,
    '<client> <nick> <account> :is logged in as',
)

RPL_NOTOPIC = _register_numeric_reply(
    'RPL_NOTOPIC',
    331,
    '<client> <channel> :No topic is set',
)

RPL_TOPIC = _register_numeric_reply(
    'RPL_TOPIC',
    332,
    '<client> <channel> :<topic>',
)

RPL_TOPICWHOTIME = _register_numeric_reply(
    'RPL_TOPICWHOTIME',
    333,
    '<client> <channel> <nick> <setat>',
)

RPL_INVITELIST = _register_numeric_reply(
    'RPL_INVITELIST',
    336,
    '<client> <channel>',
)

RPL_ENDOFINVITELIST = _register_numeric_reply(
    'RPL_ENDOFINVITELIST',
    337,
    '<client> :End of /INVITE list',
)

RPL_WHOISACTUALLY = _register_numeric_reply(
    'RPL_WHOISACTUALLY',
    338,
    '<client> <nick> :is actually ...',
    '<client> <nick> <host|ip> :Is actually using host',
    '<client> <nick> <username>@<hostname> <ip> :Is actually using host',
)

RPL_INVITING = _register_numeric_reply(
    'RPL_INVITING',
    341,
    '<client> <nick> <channel>',
)

RPL_INVEXLIST = _register_numeric_reply(
    'RPL_INVEXLIST',
    346,
    '<client> <channel> <mask>',
)

RPL_ENDOFINVEXLIST = _register_numeric_reply(
    'RPL_ENDOFINVEXLIST',
    347,
    '<client> <channel> :End of Channel Invite Exception List',
)
RPL_EXCEPTLIST = _register_numeric_reply(
    'RPL_EXCEPTLIST',
    348,
    '<client> <channel> <mask>',
)

RPL_ENDOFEXCEPTLIST = _register_numeric_reply(
    'RPL_ENDOFEXCEPTLIST',
    349,
    '<client> <channel> :End of channel exception list',
)

RPL_VERSION = _register_numeric_reply(
    'RPL_VERSION',
    351,
    '<client> <version> <server> :<comments>',
)

RPL_WHOREPLY = _register_numeric_reply(
    'RPL_WHOREPLY',
    352,
    '<client> <channel> <username> <host> <server> <nick> <flags> :<hopcount> <realname>',
)
RPL_NAMREPLY = _register_numeric_reply(
    'RPL_NAMREPLY',
    353,
    '<client> <symbol> <channel> :[prefix]<nick>{ [prefix]<nick>}',
)

RPL_LINKS = _register_numeric_reply(
    'RPL_LINKS',
    364,
    '<client> * <server> :<hopcount> <server info>',
)

RPL_ENDOFLINKS = _register_numeric_reply(
    'RPL_ENDOFLINKS',
    365,
    '<client> * :End of /LINKS list',
)

RPL_ENDOFNAMES = _register_numeric_reply(
    'RPL_ENDOFNAMES',
    366,
    '<client> <channel> :End of /NAMES list',
)

RPL_BANLIST = _register_numeric_reply(
    'RPL_BANLIST',
    367,
    '<client> <channel> <mask> [<who> <set-ts>]',
)

RPL_ENDOFBANLIST = _register_numeric_reply(
    'RPL_ENDOFBANLIST',
    368,
    '<client> <channel> :End of channel ban list',
)

RPL_ENDOFWHOWAS = _register_numeric_reply(
    'RPL_ENDOFWHOWAS',
    369,
    '<client> <nick> :End of WHOWAS',
)

RPL_INFO = _register_numeric_reply(
    'RPL_INFO',
    371,
    '<client> :<string>',
)

RPL_MOTD = _register_numeric_reply(
    'RPL_MOTD',
    372,
    '<client> :<line of the motd>',
)

RPL_ENDOFINFO = _register_numeric_reply(
    'RPL_ENDOFINFO',
    374,
    '<client> :End of INFO list',
)

RPL_MOTDSTART = _register_numeric_reply(
    'RPL_MOTDSTART',
    375,
    '<client> :- <server> Message of the day - ',
)

RPL_ENDOFMOTD = _register_numeric_reply(
    'RPL_ENDOFMOTD',
    376,
    '<client> :End of /MOTD command.',
)

RPL_WHOISHOST = _register_numeric_reply(
    'RPL_WHOISHOST',
    378,
    '<client> <nick> :is connecting from *@localhost 127.0.0.1',
)

RPL_WHOISMODES = _register_numeric_reply(
    'RPL_WHOISMODES',
    379,
    '<client> <nick> :is using modes +ailosw',
)

RPL_YOUREOPER = _register_numeric_reply(
    'RPL_YOUREOPER',
    381,
    '<client> :You are now an IRC operator',
)

RPL_REHASHING = _register_numeric_reply(
    'RPL_REHASHING',
    382,
    '<client> <config file> :Rehashing',
)

RPL_TIME = _register_numeric_reply(
    'RPL_TIME',
    391,
    '<client> <server> [<timestamp> [<TS offset>]] :<human-readable time>',
)

#

ERR_UNKNOWNERROR = _register_numeric_reply(
    'ERR_UNKNOWNERROR',
    400,
    '<client> <command>{ <subcommand>} :<info>',
)

ERR_NOSUCHNICK = _register_numeric_reply(
    'ERR_NOSUCHNICK',
    401,
    '<client> <nickname> :No such nick/channel',
)

ERR_NOSUCHSERVER = _register_numeric_reply(
    'ERR_NOSUCHSERVER',
    402,
    '<client> <server name> :No such server',
)

ERR_NOSUCHCHANNEL = _register_numeric_reply(
    'ERR_NOSUCHCHANNEL',
    403,
    '<client> <channel> :No such channel',
)

ERR_CANNOTSENDTOCHAN = _register_numeric_reply(
    'ERR_CANNOTSENDTOCHAN',
    404,
    '<client> <channel> :Cannot send to channel',
)

ERR_TOOMANYCHANNELS = _register_numeric_reply(
    'ERR_TOOMANYCHANNELS',
    405,
    '<client> <channel> :You have joined too many channels',
)

ERR_WASNOSUCHNICK = _register_numeric_reply(
    'ERR_WASNOSUCHNICK',
    406,
    '<client> <nickname> :There was no such nickname',
)

ERR_TOOMANYTARGETS = _register_numeric_reply(
    'ERR_TOOMANYTARGETS',
    407,
    '<target> :<error code> recipients. <abort message>',
)

ERR_NOORIGIN = _register_numeric_reply(
    'ERR_NOORIGIN',
    409,
    '<client> :No origin specified',
)

ERR_NORECIPIENT = _register_numeric_reply(
    'ERR_NORECIPIENT',
    411,
    '<client> :No recipient given (<command>)',
)

ERR_NOTEXTTOSEND = _register_numeric_reply(
    'ERR_NOTEXTTOSEND',
    412,
    '<client> :No text to send',
)

ERR_NOTOPLEVEL = _register_numeric_reply(
    'ERR_NOTOPLEVEL',
    413,
    '<mask> :No toplevel domain specified',
)

ERR_WILDTOPLEVEL = _register_numeric_reply(
    'ERR_WILDTOPLEVEL',
    414,
    '<mask> :Wildcard in toplevel domain',
)

ERR_INPUTTOOLONG = _register_numeric_reply(
    'ERR_INPUTTOOLONG',
    417,
    '<client> :Input line was too long',
)

ERR_UNKNOWNCOMMAND = _register_numeric_reply(
    'ERR_UNKNOWNCOMMAND',
    421,
    '<client> <command> :Unknown command',
)

ERR_NOMOTD = _register_numeric_reply(
    'ERR_NOMOTD',
    422,
    '<client> :MOTD File is missing',
)

ERR_NONICKNAMEGIVEN = _register_numeric_reply(
    'ERR_NONICKNAMEGIVEN',
    431,
    '<client> :No nickname given',
)

ERR_ERRONEUSNICKNAME = _register_numeric_reply(
    'ERR_ERRONEUSNICKNAME',
    432,
    '<client> <nick> :Erroneus nickname',
)

ERR_NICKNAMEINUSE = _register_numeric_reply(
    'ERR_NICKNAMEINUSE',
    433,
    '<client> <nick> :Nickname is already in use',
)

ERR_NICKCOLLISION = _register_numeric_reply(
    'ERR_NICKCOLLISION',
    436,
    '<client> <nick> :Nickname collision KILL from <user>@<host>',
)

ERR_USERNOTINCHANNEL = _register_numeric_reply(
    'ERR_USERNOTINCHANNEL',
    441,
    "<client> <nick> <channel> :They aren't on that channel",
)

ERR_NOTONCHANNEL = _register_numeric_reply(
    'ERR_NOTONCHANNEL',
    442,
    "<client> <channel> :You're not on that channel",
)

ERR_USERONCHANNEL = _register_numeric_reply(
    'ERR_USERONCHANNEL',
    443,
    '<client> <nick> <channel> :is already on channel',
)

ERR_NOTREGISTERED = _register_numeric_reply(
    'ERR_NOTREGISTERED',
    451,
    '<client> :You have not registered',
)

ERR_NEEDMOREPARAMS = _register_numeric_reply(
    'ERR_NEEDMOREPARAMS',
    461,
    '<client> <command> :Not enough parameters',
)

ERR_ALREADYREGISTERED = _register_numeric_reply(
    'ERR_ALREADYREGISTERED',
    462,
    '<client> :You may not reregister',
)

ERR_PASSWDMISMATCH = _register_numeric_reply(
    'ERR_PASSWDMISMATCH',
    464,
    '<client> :Password incorrect',
)

ERR_YOUREBANNEDCREEP = _register_numeric_reply(
    'ERR_YOUREBANNEDCREEP',
    465,
    '<client> :You are banned from this server.',
)

ERR_CHANNELISFULL = _register_numeric_reply(
    'ERR_CHANNELISFULL',
    471,
    '<client> <channel> :Cannot join channel (+l)',
)

ERR_UNKNOWNMODE = _register_numeric_reply(
    'ERR_UNKNOWNMODE',
    472,
    '<client> <modechar> :is unknown mode char to me',
)

ERR_INVITEONLYCHAN = _register_numeric_reply(
    'ERR_INVITEONLYCHAN',
    473,
    '<client> <channel> :Cannot join channel (+i)',
)

ERR_BANNEDFROMCHAN = _register_numeric_reply(
    'ERR_BANNEDFROMCHAN',
    474,
    '<client> <channel> :Cannot join channel (+b)',
)

ERR_BADCHANNELKEY = _register_numeric_reply(
    'ERR_BADCHANNELKEY',
    475,
    '<client> <channel> :Cannot join channel (+k)',
)

ERR_BADCHANMASK = _register_numeric_reply(
    'ERR_BADCHANMASK',
    476,
    '<channel> :Bad Channel Mask',
)

ERR_NOPRIVILEGES = _register_numeric_reply(
    'ERR_NOPRIVILEGES',
    481,
    "<client> :Permission Denied- You're not an IRC operator",
)

ERR_CHANOPRIVSNEEDED = _register_numeric_reply(
    'ERR_CHANOPRIVSNEEDED',
    482,
    "<client> <channel> :You're not channel operator",
)

ERR_CANTKILLSERVER = _register_numeric_reply(
    'ERR_CANTKILLSERVER',
    483,
    '<client> :You cant kill a server!',
)

ERR_NOOPERHOST = _register_numeric_reply(
    'ERR_NOOPERHOST',
    491,
    '<client> :No O-lines for your host',
)

ERR_NOSERVICEHOST = _register_numeric_reply('ERR_NOSERVICEHOST', 492)

ERR_UMODEUNKNOWNFLAG = _register_numeric_reply(
    'ERR_UMODEUNKNOWNFLAG',
    501,
    '<client> :Unknown MODE flag',
)

ERR_USERSDONTMATCH = _register_numeric_reply(
    'ERR_USERSDONTMATCH',
    502,
    '<client> :Cant change mode for other users',
)

ERR_HELPNOTFOUND = _register_numeric_reply(
    'ERR_HELPNOTFOUND',
    524,
    '<client> <subject> :No help available on this topic',
)

ERR_INVALIDKEY = _register_numeric_reply(
    'ERR_INVALIDKEY',
    525,
)

RPL_STARTTLS = _register_numeric_reply(
    'RPL_STARTTLS',
    670,
    '<client> :STARTTLS successful, proceed with TLS handshake',
)

RPL_WHOISSECURE = _register_numeric_reply(
    'RPL_WHOISSECURE',
    671,
    '<client> <nick> :is using a secure connection',
)

ERR_STARTTLS = _register_numeric_reply(
    'ERR_STARTTLS',
    691,
    '<client> :STARTTLS failed (Wrong moon phase)',
)

ERR_INVALIDMODEPARAM = _register_numeric_reply(
    'ERR_INVALIDMODEPARAM',
    696,
)

RPL_HELPSTART = _register_numeric_reply(
    'RPL_HELPSTART',
    704,
)

RPL_HELPTXT = _register_numeric_reply(
    'RPL_HELPTXT',
    705,
)

RPL_ENDOFHELP = _register_numeric_reply(
    'RPL_ENDOFHELP',
    706,
)

ERR_NOPRIVS = _register_numeric_reply(
    'ERR_NOPRIVS',
    723,
    '<client> <priv> :Insufficient oper privileges.',
)

RPL_LOGGEDIN = _register_numeric_reply(
    'RPL_LOGGEDIN',
    900,
    '<client> <nick>!<user>@<host> <account> :You are now logged in as <username>',
)

RPL_LOGGEDOUT = _register_numeric_reply(
    'RPL_LOGGEDOUT',
    901,
    '<client> <nick>!<user>@<host> :You are now logged out',
)

ERR_NICKLOCKED = _register_numeric_reply(
    'ERR_NICKLOCKED',
    902,
    '<client> :You must use a nick assigned to you',
)

RPL_SASLSUCCESS = _register_numeric_reply(
    'RPL_SASLSUCCESS',
    903,
    '<client> :SASL authentication successful',
)

ERR_SASLFAIL = _register_numeric_reply(
    'ERR_SASLFAIL',
    904,
    '<client> :SASL authentication failed',
)

ERR_SASLTOOLONG = _register_numeric_reply(
    'ERR_SASLTOOLONG',
    905,
    '<client> :SASL message too long',
)

ERR_SASLABORTED = _register_numeric_reply(
    'ERR_SASLABORTED',
    906,
    '<client> :SASL authentication aborted',
)

ERR_SASLALREADY = _register_numeric_reply(
    'ERR_SASLALREADY',
    907,
    '<client> :You have already authenticated using SASL',
)

RPL_SASLMECHS = _register_numeric_reply(
    'RPL_SASLMECHS',
    908,
    '<client> <mechanisms> :are available SASL mechanisms',
)


##


NUMERIC_REPLIES = NumericReplies(_REGISTERED_NUMERIC_REPLIES)



import typing as ta

from .... import dataclasses as dc
from .base import Message
from .formats import MessageFormat


##


class ParseError(Exception):
    pass


def parse_message(cls: type[Message], params: ta.Sequence[str]) -> Message:
    mf = cls.FORMAT

    kws: dict = {}
    i = 0
    for fp in mf.params:
        if isinstance(fp, MessageFormat.KwargParam):
            if i >= len(params):
                if not fp.optional:
                    raise ParseError(f'Missing param: {fp.name}')
                continue

            kv: ta.Any
            if (ar := fp.arity) is MessageFormat.KwargParam.Arity.SINGLE:
                kv = params[i]
                i += 1

            elif ar is MessageFormat.KwargParam.Arity.VARIADIC:
                kv = params[i:]
                i = len(params)

            elif ar is MessageFormat.KwargParam.Arity.COMMA_LIST:
                kv = params[i].split(',')
                i += 1

            else:
                raise TypeError(ar)

            kws[fp.name] = kv

        elif isinstance(fp, MessageFormat.LiteralParam):
            if i >= len(params):
                raise ParseError('Missing literal param')

            pv = params[i]
            if fp.text != pv:
                raise ParseError(f'Unexpected literal: {pv}')
            i += 1

        else:
            raise TypeError(fp)

    if i != len(params):
        raise ParseError('Unconsumed params')

    if (up := mf.unpack_params) is not None:
        kws = dict(up(kws))

    return cls(**kws)


##


class UnparsedMessage(ta.NamedTuple):
    name: str
    params: ta.Sequence[str]


def unparse_message(msg: Message) -> UnparsedMessage:
    mf = msg.FORMAT

    kws = {k: v for k, v in dc.asdict(msg).items() if v is not None}

    if (up := mf.unpack_params) is not None:
        kws = dict(up.backward(kws))

    params = []

    for fp in mf.params:
        if isinstance(fp, MessageFormat.KwargParam):
            # FIXME
            raise NotImplementedError

        elif isinstance(fp, MessageFormat.LiteralParam):
            params.append(fp.text)

        else:
            raise TypeError(fp)

    return UnparsedMessage(
        name=mf.name,
        params=params,
    )



# ruff: noqa
# flake8: noqa
# type: ignore
from .Token import Token
from .InputStream import InputStream
from .FileStream import FileStream
from .StdinStream import StdinStream
from .BufferedTokenStream import TokenStream
from .CommonTokenStream import CommonTokenStream
from .Lexer import Lexer
from .Parser import Parser
from .dfa.DFA import DFA
from .atn.ATN import ATN
from .atn.ATNDeserializer import ATNDeserializer
from .atn.LexerATNSimulator import LexerATNSimulator
from .atn.ParserATNSimulator import ParserATNSimulator
from .atn.PredictionMode import PredictionMode
from .PredictionContext import PredictionContextCache
from .ParserRuleContext import RuleContext, ParserRuleContext
from .tree.Tree import ParseTreeListener, ParseTreeVisitor, ParseTreeWalker, TerminalNode, ErrorNode, RuleNode
from .error.Errors import RecognitionException, IllegalStateException, NoViableAltException
from .error.ErrorStrategy import BailErrorStrategy
from .error.DiagnosticErrorListener import DiagnosticErrorListener
from .Utils import str_list



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
TODO:
 - popen
 - route check_calls through run_?
"""
import abc
import subprocess
import sys
import typing as ta

from ..lite.timeouts import TimeoutLike
from .base import BaseSubprocesses
from .run import SubprocessRun
from .run import SubprocessRunOutput


##


class AbstractSubprocesses(BaseSubprocesses, abc.ABC):
    @abc.abstractmethod
    def run_(self, run: SubprocessRun) -> SubprocessRunOutput:
        raise NotImplementedError

    def run(
            self,
            *cmd: str,
            input: ta.Any = None,  # noqa
            timeout: TimeoutLike = None,
            check: bool = False,
            capture_output: ta.Optional[bool] = None,
            **kwargs: ta.Any,
    ) -> SubprocessRunOutput:
        return self.run_(SubprocessRun(
            cmd=cmd,
            input=input,
            timeout=timeout,
            check=check,
            capture_output=capture_output,
            kwargs=kwargs,
        ))

    #

    @abc.abstractmethod
    def check_call(
            self,
            *cmd: str,
            stdout: ta.Any = sys.stderr,
            **kwargs: ta.Any,
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def check_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> bytes:
        raise NotImplementedError

    #

    def check_output_str(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> str:
        return self.check_output(*cmd, **kwargs).decode().strip()

    #

    def try_call(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> bool:
        if isinstance(self.try_fn(self.check_call, *cmd, **kwargs), Exception):
            return False
        else:
            return True

    def try_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> ta.Optional[bytes]:
        if isinstance(ret := self.try_fn(self.check_output, *cmd, **kwargs), Exception):
            return None
        else:
            return ret

    def try_output_str(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> ta.Optional[str]:
        if (ret := self.try_output(*cmd, **kwargs)) is None:
            return None
        else:
            return ret.decode().strip()


##


class Subprocesses(AbstractSubprocesses):
    def run_(self, run: SubprocessRun) -> SubprocessRunOutput[subprocess.CompletedProcess]:
        with self.prepare_and_wrap(
                *run.cmd,
                input=run.input,
                timeout=run.timeout,
                check=run.check,
                capture_output=run.capture_output or False,
                **(run.kwargs or {}),
        ) as (cmd, kwargs):
            proc = subprocess.run(cmd, **kwargs)  # noqa

        return SubprocessRunOutput(
            proc=proc,

            returncode=proc.returncode,

            stdout=proc.stdout,  # noqa
            stderr=proc.stderr,  # noqa
        )

    def check_call(
            self,
            *cmd: str,
            stdout: ta.Any = sys.stderr,
            **kwargs: ta.Any,
    ) -> None:
        with self.prepare_and_wrap(*cmd, stdout=stdout, **kwargs) as (cmd, kwargs):  # noqa
            subprocess.check_call(cmd, **kwargs)

    def check_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> bytes:
        with self.prepare_and_wrap(*cmd, **kwargs) as (cmd, kwargs):  # noqa
            return subprocess.check_output(cmd, **kwargs)


##


subprocesses = Subprocesses()

SubprocessRun._DEFAULT_SUBPROCESSES = subprocesses  # noqa



import typing as ta

from .. import check
from .. import lang
from .. import reflect as rfl
from .values import TypedValue  # noqa


TypedValueT = ta.TypeVar('TypedValueT', bound='TypedValue')


##


class TypedValueGeneric(lang.Abstract, ta.Generic[TypedValueT]):
    _typed_value_type: ta.ClassVar[rfl.Type]

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        if '_typed_value_type' in cls.__dict__:
            return

        g_mro = rfl.ALIAS_UPDATING_GENERIC_SUBSTITUTION.generic_mro(cls)
        g_tvg = check.single(gb for gb in g_mro if isinstance(gb, rfl.Generic) and gb.cls is TypedValueGeneric)
        tvt = check.single(g_tvg.args)
        cls._typed_value_type = tvt



"""
TODO:
 - cryptography vs pycryptodome[x]
 - standardize failure exception
 - chains - take first
 - keysets

See:
 - https://soatok.blog/2020/05/13/why-aes-gcm-sucks/
 - https://pycryptodome.readthedocs.io/en/latest/src/cipher/chacha20_poly1305.html

==

https://www.tecmint.com/gpg-encrypt-decrypt-files/

==

gpg --batch --passphrase '' --quick-gen-key wrmsr default default
gpg --batch --passphrase '' --quick-gen-key wrmsr2 default default
echo 'hi there' > secret.txt
gpg -e -u wrmsr -r wrmsr2 secret.txt
gpg -d -o secret2.txt secret.txt.gpg

gpg --batch -c --passphrase-file /var/secret.key -o some.gpg toencrypt.txt

openssl rand -rand /dev/urandom 128 > barf.key
openssl enc -in secret.txt -out secret.txt.enc -e -aes256 -pbkdf2 -kfile barf.key
openssl aes-256-cbc -d -pbkdf2 -in secret.txt.enc -out secret3.txt -kfile barf.key

https://wiki.openssl.org/index.php/Enc#Options
-pass 'file:...'
"""
import abc
import secrets
import typing as ta

from .. import lang


if ta.TYPE_CHECKING:
    from cryptography import exceptions as cry_exc
    from cryptography import fernet as cry_fernet
    from cryptography.hazmat.primitives.ciphers import aead as cry_aead
else:
    cry_aead = lang.proxy_import('cryptography.hazmat.primitives.ciphers.aead')
    cry_exc = lang.proxy_import('cryptography.exceptions')
    cry_fernet = lang.proxy_import('cryptography.fernet')


##


class EncryptionError(Exception):
    pass


class DecryptionError(Exception):
    pass


class Crypto(abc.ABC):
    @abc.abstractmethod
    def generate_key(self) -> bytes:
        raise NotImplementedError

    @abc.abstractmethod
    def encrypt(self, data: bytes, key: bytes) -> bytes:
        raise NotImplementedError

    @abc.abstractmethod
    def decrypt(self, data: bytes, key: bytes) -> bytes:
        raise NotImplementedError


##


class FernetCrypto(Crypto):
    def generate_key(self) -> bytes:
        return cry_fernet.Fernet.generate_key()

    def encrypt(self, data: bytes, key: bytes) -> bytes:
        try:
            f = cry_fernet.Fernet(key)
        except ValueError as e:
            raise EncryptionError from e
        return f.encrypt(data)

    def decrypt(self, data: bytes, key: bytes) -> bytes:
        try:
            f = cry_fernet.Fernet(key)
        except ValueError as e:
            raise DecryptionError from e
        try:
            return f.decrypt(data)
        except cry_fernet.InvalidToken as e:
            raise DecryptionError from e


class AesgcmCrypto(Crypto):
    """https://stackoverflow.com/a/59835994"""

    def generate_key(self) -> bytes:
        return secrets.token_bytes(32)

    def encrypt(self, data: bytes, key: bytes) -> bytes:
        nonce = secrets.token_bytes(12)
        return nonce + cry_aead.AESGCM(key).encrypt(nonce, data, b'')

    def decrypt(self, data: bytes, key: bytes) -> bytes:
        try:
            return cry_aead.AESGCM(key).decrypt(data[:12], data[12:], b'')
        except cry_exc.InvalidTag as e:
            raise DecryptionError from e


class Chacha20Poly1305Crypto(Crypto):
    """https://cryptography.io/en/latest/hazmat/primitives/aead/#cryptography.hazmat.primitives.ciphers.aead.ChaCha20Poly1305"""  # noqa

    def generate_key(self) -> bytes:
        return cry_aead.ChaCha20Poly1305.generate_key()

    def encrypt(self, data: bytes, key: bytes) -> bytes:
        nonce = secrets.token_bytes(12)
        return nonce + cry_aead.ChaCha20Poly1305(key).encrypt(nonce, data, b'')

    def decrypt(self, data: bytes, key: bytes) -> bytes:
        try:
            return cry_aead.ChaCha20Poly1305(key).decrypt(data[:12], data[12:], b'')
        except cry_exc.InvalidTag as e:
            raise DecryptionError from e



"""
TODO:
 - stream
  - chunk size - httpx interface is awful, punch through?
 - httpx catch
 - return non-200 HttpResponses
 - async
"""
import abc
import contextlib
import typing as ta

from ... import cached
from ... import dataclasses as dc
from ... import lang
from ..headers import CanHttpHeaders
from ..headers import HttpHeaders


BaseHttpResponseT = ta.TypeVar('BaseHttpResponseT', bound='BaseHttpResponse')


##


DEFAULT_ENCODING = 'utf-8'


def is_success_status(status: int) -> bool:
    return 200 <= status < 300


##


@dc.dataclass(frozen=True)
class HttpRequest(lang.Final):
    url: str
    method: str | None = None  # noqa

    _: dc.KW_ONLY

    headers: CanHttpHeaders | None = dc.xfield(None, repr=dc.truthy_repr)
    data: bytes | str | None = dc.xfield(None, repr_fn=lambda v: '...' if v is not None else None)

    timeout_s: float | None = None

    #

    @property
    def method_or_default(self) -> str:
        if self.method is not None:
            return self.method
        if self.data is not None:
            return 'POST'
        return 'GET'

    @cached.property
    def headers_(self) -> HttpHeaders | None:
        return HttpHeaders(self.headers) if self.headers is not None else None


#


@dc.dataclass(frozen=True, kw_only=True)
class BaseHttpResponse(lang.Abstract):
    status: int

    headers: HttpHeaders | None = dc.xfield(None, repr=dc.truthy_repr)

    request: HttpRequest
    underlying: ta.Any = dc.field(default=None, repr=False)

    @property
    def is_success(self) -> bool:
        return is_success_status(self.status)


@dc.dataclass(frozen=True, kw_only=True)
class HttpResponse(BaseHttpResponse, lang.Final):
    data: bytes | None = dc.xfield(None, repr_fn=lambda v: '...' if v is not None else None)


@dc.dataclass(frozen=True, kw_only=True)
class StreamHttpResponse(BaseHttpResponse, lang.Final):
    class Stream(ta.Protocol):
        def read(self, /, n: int = -1) -> bytes: ...

    stream: Stream

    _closer: ta.Callable[[], None] | None = dc.field(default=None, repr=False)

    def __enter__(self) -> ta.Self:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self) -> None:
        if (c := self._closer) is not None:
            c()


def close_response(resp: BaseHttpResponse) -> None:
    if isinstance(resp, HttpResponse):
        pass

    elif isinstance(resp, StreamHttpResponse):
        resp.close()

    else:
        raise TypeError(resp)


@contextlib.contextmanager
def closing_response(resp: BaseHttpResponseT) -> ta.Iterator[BaseHttpResponseT]:
    if isinstance(resp, HttpResponse):
        yield resp
        return

    elif isinstance(resp, StreamHttpResponse):
        with contextlib.closing(resp):
            yield resp

    else:
        raise TypeError(resp)


def read_response(resp: BaseHttpResponse) -> HttpResponse:
    if isinstance(resp, HttpResponse):
        return resp

    elif isinstance(resp, StreamHttpResponse):
        data = resp.stream.read()
        return HttpResponse(**{
            **{k: v for k, v in dc.shallow_asdict(resp).items() if k not in ('stream', '_closer')},
            'data': data,
        })

    else:
        raise TypeError(resp)


#


class HttpClientError(Exception):
    @property
    def cause(self) -> BaseException | None:
        return self.__cause__


@dc.dataclass(frozen=True)
class HttpStatusError(HttpClientError):
    response: HttpResponse


#


class HttpClient(lang.Abstract):
    def __enter__(self) -> ta.Self:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def request(
            self,
            req: HttpRequest,
            *,
            check: bool = False,
    ) -> HttpResponse:
        with closing_response(self.stream_request(
            req,
            check=check,
        )) as resp:
            return read_response(resp)

    def stream_request(
            self,
            req: HttpRequest,
            *,
            check: bool = False,
    ) -> StreamHttpResponse:
        resp = self._stream_request(req)

        try:
            if check and not resp.is_success:
                if isinstance(resp.underlying, Exception):
                    cause = resp.underlying
                else:
                    cause = None
                raise HttpStatusError(read_response(resp)) from cause  # noqa

        except Exception:
            close_response(resp)
            raise

        return resp

    @abc.abstractmethod
    def _stream_request(self, req: HttpRequest) -> StreamHttpResponse:
        raise NotImplementedError



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import contextlib
import errno
import fcntl
import os
import typing as ta


##


def is_fd_open(fd: int) -> bool:
    try:
        fcntl.fcntl(fd, fcntl.F_GETFD)
    except OSError as e:
        if e.errno == errno.EBADF:
            return False
        raise
    else:
        return True


def touch(path: str, mode: int = 0o666, exist_ok: bool = True) -> None:
    if exist_ok:
        # First try to bump modification time
        # Implementation note: GNU touch uses the UTIME_NOW option of the utimensat() / futimens() functions.
        try:
            os.utime(path, None)
        except OSError:
            pass
        else:
            return

    flags = os.O_CREAT | os.O_WRONLY
    if not exist_ok:
        flags |= os.O_EXCL

    fd = os.open(path, flags, mode)
    os.close(fd)


def unlink_if_exists(path: str) -> None:
    try:
        os.unlink(path)
    except FileNotFoundError:
        pass


@contextlib.contextmanager
def unlinking_if_exists(path: str) -> ta.Iterator[None]:
    try:
        yield
    finally:
        unlink_if_exists(path)



"""
TODO:
 - .super(instance[_cls], owner)
 - ALT: A.f(super(), ... ? :/
 - classmethod/staticmethod
"""
import contextlib
import functools
import typing as ta
import weakref

from .. import check
from .. import lang
from .dispatch import Dispatcher
from .impls import get_impl_func_cls_set


T = ta.TypeVar('T')
R = ta.TypeVar('R')
P = ta.ParamSpec('P')


##


class Method(ta.Generic[P, R]):
    """
    MRO-honoring instancemethod singledispatch. There are many ways to do this, and this one is attr name based: a class
    is considered to have method implementations that have been registered to a given method based on whether or not
    they are accessible by a non-shadowed, MRO-resolved named attribute on that class.

    Care must be taken when overriding registered implementations from superclasses in subclasses - shadowing the name
    of the superclass method will not automatically register the new method with the same name to the dispatch method -
    it must be explicitly `@register`'ed itself. This is a feature, allowing for selective de-registration of
    implementations in subclasses via name shadowing.

    Methods have ability to choose to allow external installation of implementations outside of direct subclasses. This
    is to be used *extremely* rarely - basically only in the rare case of externally extensible type hierarchies with
    visitors.
    """

    def __init__(
            self,
            func: ta.Callable,
            *,
            installable: bool = False,
            requires_override: bool = False,
    ) -> None:
        super().__init__()

        if not callable(func) and not hasattr(func, '__get__'):  # type: ignore
            raise TypeError(f'{func!r} is not callable or a descriptor')

        self._func = func
        self._installable = installable
        self._requires_override = requires_override

        self._impls: ta.MutableMapping[ta.Callable, frozenset[type] | None] = weakref.WeakKeyDictionary()

        # bpo-45678: special-casing for classmethod/staticmethod in Python <=3.9, as functools.update_wrapper doesn't
        # work properly in singledispatchmethod.__get__ if it is applied to an unbound classmethod/staticmethod
        unwrapped_func: ta.Any
        if isinstance(func, (staticmethod, classmethod)):
            unwrapped_func = func.__func__
        else:
            unwrapped_func = func
        self._unwrapped_func = unwrapped_func
        self._is_abstractmethod = getattr(func, '__isabstractmethod__', False)  # noqa
        self.update_wrapper(self)

        self._dispatch_func_cache: dict[ta.Any, ta.Callable] = {}

        def dispatch_func_cache_remove(k, self_ref=weakref.ref(self)):
            if (ref_self := self_ref()) is not None:
                cache = ref_self._dispatch_func_cache  # noqa
                with contextlib.suppress(KeyError):
                    del cache[k]

        self._dispatch_func_cache_remove = dispatch_func_cache_remove

        self._owner: type | None = None
        self._name: str | None = None

    def __set_name__(self, owner, name):
        if self._owner is None:
            self._owner = owner
        if self._name is None:
            self._name = name

    def __repr__(self) -> str:
        return f'<{type(self).__module__}.{type(self).__qualname__}:{self._func} at 0x{id(self):x}>'

    def update_wrapper(self, wrapper: T) -> T:
        for attr in functools.WRAPPER_ASSIGNMENTS:
            try:
                value = getattr(self._unwrapped_func, attr)
            except AttributeError:
                continue
            setattr(wrapper, attr, value)
        setattr(wrapper, '__isabstractmethod__', self._is_abstractmethod)  # noqa
        return wrapper

    def register(self, impl: T, cls_set: frozenset[type] | None = None) -> T:
        # bpo-39679: in Python <= 3.9, classmethods and staticmethods don't inherit __annotations__ of the wrapped
        # function (fixed in 3.10+ as a side-effect of bpo-43682) but we need that for annotation-derived
        # singledispatches. So we add that just-in-time here.
        if isinstance(impl, (staticmethod, classmethod)):
            impl.__annotations__ = getattr(impl.__func__, '__annotations__', {})

        check.callable(impl)
        if impl not in self._impls:
            self._impls[impl] = cls_set  # type: ignore
            self._dispatch_func_cache.clear()

        return impl

    def _is_impl(self, obj: ta.Any) -> bool:
        try:
            hash(obj)
        except TypeError:
            return False

        return obj in self._impls

    def build_attr_dispatcher(self, instance_cls: type, owner_cls: type | None = None) -> Dispatcher[str]:
        if owner_cls is None:
            owner_cls = instance_cls

        mro = instance_cls.__mro__[-2::-1]
        try:
            mro_pos = mro.index(owner_cls)
        except ValueError:
            raise TypeError(f'Owner class {owner_cls} not in mro of instance class {instance_cls}') from None

        mro_dct: dict[str, list[tuple[type, ta.Any]]] = {}
        for cur_cls in mro[:mro_pos + 1]:
            for att, obj in cur_cls.__dict__.items():
                if att not in mro_dct:
                    if not self._is_impl(obj):
                        continue

                try:
                    lst = mro_dct[att]
                except KeyError:
                    lst = mro_dct[att] = []
                lst.append((cur_cls, obj))

        #

        disp: Dispatcher[str] = Dispatcher()

        seen: dict[ta.Any, str] = {}
        for att, lst in mro_dct.items():
            if not lst:
                raise RuntimeError
            _, obj = lst[-1]

            if len(lst) > 1:
                if self._requires_override and not lang.is_override(obj):
                    raise lang.RequiresOverrideError(
                        att,
                        instance_cls,
                        lst[-1][0],
                        lst[0][0],
                    )

            if not self._is_impl(obj):
                continue

            cls_set = self._impls[obj]
            if cls_set is None:
                cls_set = get_impl_func_cls_set(obj, arg_offset=1)
                self._impls[obj] = cls_set

            try:
                ex_att = seen[obj]
            except KeyError:
                pass
            else:
                raise TypeError(f'Duplicate impl: {owner_cls} {instance_cls} {att} {ex_att}')
            seen[obj] = att

            disp.register(att, cls_set)

        return disp

    def build_dispatch_func(self, disp: Dispatcher[str]) -> ta.Callable:
        dispatch = disp.dispatch
        type_ = type
        getattr_ = getattr
        base_func = self._func
        func_name = getattr(base_func, '__name__', 'singledispatch method')

        def __call__(self, *args, **kwargs):  # noqa
            if not args:
                raise TypeError(f'{func_name} requires at least 1 positional argument')

            if (impl_att := dispatch(type_(args[0]))) is not None:
                fn = getattr_(self, impl_att)
                return fn(*args, **kwargs)

            return base_func.__get__(self)(*args, **kwargs)  # noqa

        self.update_wrapper(__call__)
        return __call__

    def get_dispatch_func(self, instance_cls: type) -> ta.Callable:
        cls_ref = weakref.ref(instance_cls)
        try:
            return self._dispatch_func_cache[cls_ref]
        except KeyError:
            pass
        del cls_ref

        att_disp = self.build_attr_dispatcher(instance_cls)
        func = self.build_dispatch_func(att_disp)
        self._dispatch_func_cache[weakref.ref(instance_cls, self._dispatch_func_cache_remove)] = func
        return func

    def __get__(self, instance, owner=None):
        if instance is None:
            # FIXME: classmethod/staticmethod
            return self

        instance_cls = type(instance)
        try:
            func = self._dispatch_func_cache[weakref.ref(instance_cls)]
        except KeyError:
            func = self.get_dispatch_func(instance_cls)
        return func.__get__(instance, owner)  # noqa

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        instance, *rest = args
        instance_cls = type(instance)

        # if instance_cls is super:
        #     owner = instance.__self_class__.__mro__[instance.__self_class__.__mro__.index(instance.__thisclass__) + 1]
        #     att_disp = self.build_attr_dispatcher(instance.__self_class__, owner)
        #     func = self.build_dispatch_func(att_disp)
        #     return func.__get__(instance, instance.__thisclass__)(*rest, **kwargs)

        try:
            func = self._dispatch_func_cache[weakref.ref(instance_cls)]
        except KeyError:
            func = self.get_dispatch_func(instance_cls)
        return func.__get__(instance)(*rest, **kwargs)  # noqa


##


@ta.overload
def method(
        func: ta.Callable[P, R],
        /,
        *,
        installable: bool = False,
        requires_override: bool = False,
) -> Method[P, R]:  # noqa
    ...


@ta.overload
def method(
        func: None = None,
        /,
        *,
        installable: bool = False,
        requires_override: bool = False,
) -> ta.Callable[[ta.Callable[P, R]], Method[P, R]]:  # noqa
    ...


def method(
        func=None,
        /,
        *,
        installable=False,
        requires_override=False,
):
    kw = dict(
        installable=installable,
        requires_override=requires_override,
    )

    if func is None:
        return functools.partial(Method, **kw)

    return Method(func, **kw)


#


def install_method(
        mth: ta.Any,
        *,
        name: str | None = None,
        on: type | None = None,
) -> ta.Callable[[T], T]:
    mth = check.isinstance(mth, Method)
    if not mth._installable:  # noqa
        raise TypeError(f'Method not installable: {mth}')

    def inner(fn):
        a = name
        if a is None:
            a = fn.__name__

        owner: type = check.not_none(mth._owner)  # noqa
        if on is None:
            cls = owner
        else:
            cls = check.issubclass(on, owner)

        check.arg(not hasattr(cls, a))
        setattr(cls, a, fn)

        mth.register(fn)

        return fn

    return inner



"""
See:
 - https://github.com/oremanj/trio-monitor
 - https://github.com/python-trio/trio-monitor/tree/master
 - https://github.com/python-trio/triopg
 - https://github.com/python-trio/trio-mysql
 - https://github.com/goodboy/tractor
 - https://github.com/groove-x/trio-util
 - https://github.com/oremanj/tricycle/tree/master
 - https://github.com/linkdd/triotp
"""



"""
TODO:
 - pickle stdlib objs? have to pin to 3.8 pickle protocol, will be cross-version
 - Options.sequence_cls = list, mapping_cls = dict, ... - def with_mutable_containers() -> Options
"""
# ruff: noqa: UP006 UP007 UP045
import abc
import base64
import collections.abc
import dataclasses as dc  # noqa
import datetime
import decimal
import enum
import fractions
import functools
import inspect
import threading
import typing as ta
import uuid
import weakref

from .check import check
from .reflect import deep_subclasses
from .reflect import get_literal_type_args
from .reflect import get_new_type_supertype
from .reflect import is_generic_alias
from .reflect import is_literal_type
from .reflect import is_new_type
from .reflect import is_union_alias
from .strings import snake_case


T = ta.TypeVar('T')


##


@dc.dataclass(frozen=True)
class ObjMarshalOptions:
    raw_bytes: bool = False
    non_strict_fields: bool = False


class ObjMarshaler(abc.ABC):
    @abc.abstractmethod
    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        raise NotImplementedError

    @abc.abstractmethod
    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        raise NotImplementedError


class NopObjMarshaler(ObjMarshaler):
    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return o

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return o


@dc.dataclass()
class ProxyObjMarshaler(ObjMarshaler):
    m: ta.Optional[ObjMarshaler] = None

    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return check.not_none(self.m).marshal(o, ctx)

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return check.not_none(self.m).unmarshal(o, ctx)


@dc.dataclass(frozen=True)
class CastObjMarshaler(ObjMarshaler):
    ty: type

    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return o

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return self.ty(o)


class DynamicObjMarshaler(ObjMarshaler):
    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return ctx.manager.marshal_obj(o, opts=ctx.options)

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return o


@dc.dataclass(frozen=True)
class Base64ObjMarshaler(ObjMarshaler):
    ty: type

    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return base64.b64encode(o).decode('ascii')

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return self.ty(base64.b64decode(o))


@dc.dataclass(frozen=True)
class BytesSwitchedObjMarshaler(ObjMarshaler):
    m: ObjMarshaler

    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        if ctx.options.raw_bytes:
            return o
        return self.m.marshal(o, ctx)

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        if ctx.options.raw_bytes:
            return o
        return self.m.unmarshal(o, ctx)


@dc.dataclass(frozen=True)
class EnumObjMarshaler(ObjMarshaler):
    ty: type

    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return o.name

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return self.ty.__members__[o]  # type: ignore


@dc.dataclass(frozen=True)
class OptionalObjMarshaler(ObjMarshaler):
    item: ObjMarshaler

    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        if o is None:
            return None
        return self.item.marshal(o, ctx)

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        if o is None:
            return None
        return self.item.unmarshal(o, ctx)


@dc.dataclass(frozen=True)
class PrimitiveUnionObjMarshaler(ObjMarshaler):
    pt: ta.Tuple[type, ...]
    x: ta.Optional[ObjMarshaler] = None

    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        if isinstance(o, self.pt):
            return o
        elif self.x is not None:
            return self.x.marshal(o, ctx)
        else:
            raise TypeError(o)

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        if isinstance(o, self.pt):
            return o
        elif self.x is not None:
            return self.x.unmarshal(o, ctx)
        else:
            raise TypeError(o)


@dc.dataclass(frozen=True)
class LiteralObjMarshaler(ObjMarshaler):
    item: ObjMarshaler
    vs: frozenset

    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return self.item.marshal(check.in_(o, self.vs), ctx)

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return check.in_(self.item.unmarshal(o, ctx), self.vs)


@dc.dataclass(frozen=True)
class MappingObjMarshaler(ObjMarshaler):
    ty: type
    km: ObjMarshaler
    vm: ObjMarshaler

    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return {self.km.marshal(k, ctx): self.vm.marshal(v, ctx) for k, v in o.items()}

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return self.ty((self.km.unmarshal(k, ctx), self.vm.unmarshal(v, ctx)) for k, v in o.items())


@dc.dataclass(frozen=True)
class IterableObjMarshaler(ObjMarshaler):
    ty: type
    item: ObjMarshaler

    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return [self.item.marshal(e, ctx) for e in o]

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return self.ty(self.item.unmarshal(e, ctx) for e in o)


@dc.dataclass(frozen=True)
class FieldsObjMarshaler(ObjMarshaler):
    ty: type

    @dc.dataclass(frozen=True)
    class Field:
        att: str
        key: str
        m: ObjMarshaler

        omit_if_none: bool = False

    fs: ta.Sequence[Field]

    non_strict: bool = False

    #

    _fs_by_att: ta.ClassVar[ta.Mapping[str, Field]]
    _fs_by_key: ta.ClassVar[ta.Mapping[str, Field]]

    def __post_init__(self) -> None:
        fs_by_att: dict = {}
        fs_by_key: dict = {}
        for f in self.fs:
            check.not_in(check.non_empty_str(f.att), fs_by_att)
            check.not_in(check.non_empty_str(f.key), fs_by_key)
            fs_by_att[f.att] = f
            fs_by_key[f.key] = f
        self.__dict__['_fs_by_att'] = fs_by_att
        self.__dict__['_fs_by_key'] = fs_by_key

    #

    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        d = {}
        for f in self.fs:
            mv = f.m.marshal(getattr(o, f.att), ctx)
            if mv is None and f.omit_if_none:
                continue
            d[f.key] = mv
        return d

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        kw = {}
        for k, v in o.items():
            if (f := self._fs_by_key.get(k)) is None:
                if not (self.non_strict or ctx.options.non_strict_fields):
                    raise KeyError(k)
                continue
            kw[f.att] = f.m.unmarshal(v, ctx)
        return self.ty(**kw)


@dc.dataclass(frozen=True)
class SingleFieldObjMarshaler(ObjMarshaler):
    ty: type
    fld: str

    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return getattr(o, self.fld)

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return self.ty(**{self.fld: o})


@dc.dataclass(frozen=True)
class PolymorphicObjMarshaler(ObjMarshaler):
    class Impl(ta.NamedTuple):
        ty: type
        tag: str
        m: ObjMarshaler

    impls_by_ty: ta.Mapping[type, Impl]
    impls_by_tag: ta.Mapping[str, Impl]

    @classmethod
    def of(cls, impls: ta.Iterable[Impl]) -> 'PolymorphicObjMarshaler':
        return cls(
            {i.ty: i for i in impls},
            {i.tag: i for i in impls},
        )

    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        impl = self.impls_by_ty[type(o)]
        return {impl.tag: impl.m.marshal(o, ctx)}

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        [(t, v)] = o.items()
        impl = self.impls_by_tag[t]
        return impl.m.unmarshal(v, ctx)


@dc.dataclass(frozen=True)
class DatetimeObjMarshaler(ObjMarshaler):
    ty: type

    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return o.isoformat()

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return self.ty.fromisoformat(o)  # type: ignore


class DecimalObjMarshaler(ObjMarshaler):
    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return str(check.isinstance(o, decimal.Decimal))

    def unmarshal(self, v: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return decimal.Decimal(check.isinstance(v, str))


class FractionObjMarshaler(ObjMarshaler):
    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        fr = check.isinstance(o, fractions.Fraction)
        return [fr.numerator, fr.denominator]

    def unmarshal(self, v: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        num, denom = check.isinstance(v, list)
        return fractions.Fraction(num, denom)


class UuidObjMarshaler(ObjMarshaler):
    def marshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return str(o)

    def unmarshal(self, o: ta.Any, ctx: 'ObjMarshalContext') -> ta.Any:
        return uuid.UUID(o)


##


_DEFAULT_OBJ_MARSHALERS: ta.Dict[ta.Any, ObjMarshaler] = {
    **{t: NopObjMarshaler() for t in (type(None),)},
    **{t: CastObjMarshaler(t) for t in (int, float, str, bool)},
    **{t: BytesSwitchedObjMarshaler(Base64ObjMarshaler(t)) for t in (bytes, bytearray)},
    **{t: IterableObjMarshaler(t, DynamicObjMarshaler()) for t in (list, tuple, set, frozenset)},
    **{t: MappingObjMarshaler(t, DynamicObjMarshaler(), DynamicObjMarshaler()) for t in (dict,)},

    **{t: DynamicObjMarshaler() for t in (ta.Any, object)},

    **{t: DatetimeObjMarshaler(t) for t in (datetime.date, datetime.time, datetime.datetime)},
    decimal.Decimal: DecimalObjMarshaler(),
    fractions.Fraction: FractionObjMarshaler(),
    uuid.UUID: UuidObjMarshaler(),
}

_OBJ_MARSHALER_GENERIC_MAPPING_TYPES: ta.Dict[ta.Any, type] = {
    **{t: t for t in (dict,)},
    **{t: dict for t in (collections.abc.Mapping, collections.abc.MutableMapping)},  # noqa
}

_OBJ_MARSHALER_GENERIC_ITERABLE_TYPES: ta.Dict[ta.Any, type] = {
    **{t: t for t in (list, tuple, set, frozenset)},
    collections.abc.Set: frozenset,
    collections.abc.MutableSet: set,
    collections.abc.Sequence: tuple,
    collections.abc.MutableSequence: list,
}

_OBJ_MARSHALER_PRIMITIVE_TYPES: ta.Set[type] = {
    int,
    float,
    bool,
    str,
}


##


_REGISTERED_OBJ_MARSHALERS_BY_TYPE: ta.MutableMapping[type, ObjMarshaler] = weakref.WeakKeyDictionary()


def register_type_obj_marshaler(ty: type, om: ObjMarshaler) -> None:
    _REGISTERED_OBJ_MARSHALERS_BY_TYPE[ty] = om


def register_single_field_type_obj_marshaler(fld, ty=None):
    def inner(ty):  # noqa
        register_type_obj_marshaler(ty, SingleFieldObjMarshaler(ty, fld))
        return ty

    if ty is not None:
        return inner(ty)
    else:
        return inner


##


class ObjMarshalerFieldMetadata:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError


class OBJ_MARSHALER_FIELD_KEY(ObjMarshalerFieldMetadata):  # noqa
    pass


class OBJ_MARSHALER_OMIT_IF_NONE(ObjMarshalerFieldMetadata):  # noqa
    pass


##


class ObjMarshalerManager:
    def __init__(
            self,
            *,
            default_options: ObjMarshalOptions = ObjMarshalOptions(),

            default_obj_marshalers: ta.Dict[ta.Any, ObjMarshaler] = _DEFAULT_OBJ_MARSHALERS,  # noqa
            generic_mapping_types: ta.Dict[ta.Any, type] = _OBJ_MARSHALER_GENERIC_MAPPING_TYPES,  # noqa
            generic_iterable_types: ta.Dict[ta.Any, type] = _OBJ_MARSHALER_GENERIC_ITERABLE_TYPES,  # noqa

            registered_obj_marshalers: ta.Mapping[type, ObjMarshaler] = _REGISTERED_OBJ_MARSHALERS_BY_TYPE,
    ) -> None:
        super().__init__()

        self._default_options = default_options

        self._obj_marshalers = dict(default_obj_marshalers)
        self._generic_mapping_types = generic_mapping_types
        self._generic_iterable_types = generic_iterable_types
        self._registered_obj_marshalers = registered_obj_marshalers

        self._lock = threading.RLock()
        self._marshalers: ta.Dict[ta.Any, ObjMarshaler] = dict(_DEFAULT_OBJ_MARSHALERS)
        self._proxies: ta.Dict[ta.Any, ProxyObjMarshaler] = {}

    #

    def make_obj_marshaler(
            self,
            ty: ta.Any,
            rec: ta.Callable[[ta.Any], ObjMarshaler],
            *,
            non_strict_fields: bool = False,
    ) -> ObjMarshaler:
        if isinstance(ty, type):
            if (reg := self._registered_obj_marshalers.get(ty)) is not None:
                return reg

            if abc.ABC in ty.__bases__:
                tn = ty.__name__
                impls: ta.List[ta.Tuple[type, str]] = [  # type: ignore[var-annotated]
                    (ity, ity.__name__)
                    for ity in deep_subclasses(ty)
                    if abc.ABC not in ity.__bases__
                ]

                if all(itn.endswith(tn) for _, itn in impls):
                    impls = [
                        (ity, snake_case(itn[:-len(tn)]))
                        for ity, itn in impls
                    ]

                dupe_tns = sorted(
                    dn
                    for dn, dc in collections.Counter(itn for _, itn in impls).items()
                    if dc > 1
                )
                if dupe_tns:
                    raise KeyError(f'Duplicate impl names for {ty}: {dupe_tns}')

                return PolymorphicObjMarshaler.of([
                    PolymorphicObjMarshaler.Impl(
                        ity,
                        itn,
                        rec(ity),
                    )
                    for ity, itn in impls
                ])

            if issubclass(ty, enum.Enum):
                return EnumObjMarshaler(ty)

            if dc.is_dataclass(ty):
                return FieldsObjMarshaler(
                    ty,
                    [
                        FieldsObjMarshaler.Field(
                            att=f.name,
                            key=check.non_empty_str(fk),
                            m=rec(f.type),
                            omit_if_none=check.isinstance(f.metadata.get(OBJ_MARSHALER_OMIT_IF_NONE, False), bool),
                        )
                        for f in dc.fields(ty)
                        if (fk := f.metadata.get(OBJ_MARSHALER_FIELD_KEY, f.name)) is not None
                    ],
                    non_strict=non_strict_fields,
                )

            if issubclass(ty, tuple) and hasattr(ty, '_fields'):
                return FieldsObjMarshaler(
                    ty,
                    [
                        FieldsObjMarshaler.Field(
                            att=p.name,
                            key=p.name,
                            m=rec(p.annotation),
                        )
                        for p in inspect.signature(ty).parameters.values()
                    ],
                    non_strict=non_strict_fields,
                )

        if is_new_type(ty):
            return rec(get_new_type_supertype(ty))

        if is_literal_type(ty):
            lvs = frozenset(get_literal_type_args(ty))
            if None in lvs:
                is_opt = True
                lvs -= frozenset([None])
            else:
                is_opt = False
            lty = check.single(set(map(type, lvs)))
            lm: ObjMarshaler = LiteralObjMarshaler(rec(lty), lvs)
            if is_opt:
                lm = OptionalObjMarshaler(lm)
            return lm

        if is_generic_alias(ty):
            try:
                mt = self._generic_mapping_types[ta.get_origin(ty)]
            except KeyError:
                pass
            else:
                k, v = ta.get_args(ty)
                return MappingObjMarshaler(mt, rec(k), rec(v))

            try:
                st = self._generic_iterable_types[ta.get_origin(ty)]
            except KeyError:
                pass
            else:
                [e] = ta.get_args(ty)
                return IterableObjMarshaler(st, rec(e))

            if is_union_alias(ty):
                uts = frozenset(ta.get_args(ty))
                if None in uts or type(None) in uts:
                    is_opt = True
                    uts = frozenset(ut for ut in uts if ut not in (None, type(None)))
                else:
                    is_opt = False

                um: ObjMarshaler
                if not uts:
                    raise TypeError(ty)
                elif len(uts) == 1:
                    um = rec(check.single(uts))
                else:
                    pt = tuple({ut for ut in uts if ut in _OBJ_MARSHALER_PRIMITIVE_TYPES})
                    np_uts = {ut for ut in uts if ut not in _OBJ_MARSHALER_PRIMITIVE_TYPES}
                    if not np_uts:
                        um = PrimitiveUnionObjMarshaler(pt)
                    elif len(np_uts) == 1:
                        um = PrimitiveUnionObjMarshaler(pt, x=rec(check.single(np_uts)))
                    else:
                        raise TypeError(ty)

                if is_opt:
                    um = OptionalObjMarshaler(um)
                return um

        raise TypeError(ty)

    #

    def set_obj_marshaler(
            self,
            ty: ta.Any,
            m: ObjMarshaler,
            *,
            override: bool = False,
    ) -> None:
        with self._lock:
            if not override and ty in self._obj_marshalers:
                raise KeyError(ty)
            self._obj_marshalers[ty] = m

    def get_obj_marshaler(
            self,
            ty: ta.Any,
            *,
            no_cache: bool = False,
            **kwargs: ta.Any,
    ) -> ObjMarshaler:
        with self._lock:
            if not no_cache:
                try:
                    return self._obj_marshalers[ty]
                except KeyError:
                    pass

            try:
                return self._proxies[ty]
            except KeyError:
                pass

            rec = functools.partial(
                self.get_obj_marshaler,
                no_cache=no_cache,
                **kwargs,
            )

            p = ProxyObjMarshaler()
            self._proxies[ty] = p
            try:
                m = self.make_obj_marshaler(ty, rec, **kwargs)
            finally:
                del self._proxies[ty]
            p.m = m

            if not no_cache:
                self._obj_marshalers[ty] = m
            return m

    #

    def _make_context(self, opts: ta.Optional[ObjMarshalOptions]) -> 'ObjMarshalContext':
        return ObjMarshalContext(
            options=opts or self._default_options,
            manager=self,
        )

    def marshal_obj(
            self,
            o: ta.Any,
            ty: ta.Any = None,
            opts: ta.Optional[ObjMarshalOptions] = None,
    ) -> ta.Any:
        m = self.get_obj_marshaler(ty if ty is not None else type(o))
        return m.marshal(o, self._make_context(opts))

    def unmarshal_obj(
            self,
            o: ta.Any,
            ty: ta.Union[ta.Type[T], ta.Any],
            opts: ta.Optional[ObjMarshalOptions] = None,
    ) -> T:
        m = self.get_obj_marshaler(ty)
        return m.unmarshal(o, self._make_context(opts))

    def roundtrip_obj(
            self,
            o: ta.Any,
            ty: ta.Any = None,
            opts: ta.Optional[ObjMarshalOptions] = None,
    ) -> ta.Any:
        if ty is None:
            ty = type(o)
        m: ta.Any = self.marshal_obj(o, ty, opts)
        u: ta.Any = self.unmarshal_obj(m, ty, opts)
        return u


@dc.dataclass(frozen=True)
class ObjMarshalContext:
    options: ObjMarshalOptions
    manager: ObjMarshalerManager


##


OBJ_MARSHALER_MANAGER = ObjMarshalerManager()

set_obj_marshaler = OBJ_MARSHALER_MANAGER.set_obj_marshaler
get_obj_marshaler = OBJ_MARSHALER_MANAGER.get_obj_marshaler

marshal_obj = OBJ_MARSHALER_MANAGER.marshal_obj
unmarshal_obj = OBJ_MARSHALER_MANAGER.unmarshal_obj



"""
TODO:
 - SCOPED - eagers for EACH SCOPE
"""
from .. import check
from .. import dataclasses as dc
from .. import lang
from .elements import Element
from .keys import Key


##


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class Eager(Element, lang.Final):
    key: Key = dc.xfield(coerce=check.of_isinstance(Key))



import abc
import typing as ta

from .. import dataclasses as dc
from .. import lang
from .base import ModAttrManifest


##


class StaticModAttrManifest(dc.Static, ModAttrManifest, abc.ABC):
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        if (
                not (lang.is_abstract_class(cls) or abc.ABC in cls.__bases__) and
                'mod_name' not in cls.__dict__
        ):
            setattr(cls, 'mod_name', cls.__module__)

        super().__init_subclass__(**kwargs)



import enum
import functools
import typing as ta

from .. import check
from .. import lang


FixedWidthIntT = ta.TypeVar('FixedWidthIntT', bound='FixedWidthInt')


##


class CheckedFixedWidthIntError(ValueError):
    pass


class OverflowFixedWidthIntError(CheckedFixedWidthIntError):
    pass


class UnderflowFixedWidthIntError(CheckedFixedWidthIntError):
    pass


##


def _get_exclusive_base_cls(cls: type, base_classes: ta.Iterable[type]) -> type:
    return check.single(bcls for bcls in base_classes if issubclass(cls, bcls))


def _gen_scalar_proxy_method(name):
    def inner(self, *args, **kwargs):
        return self.__class__(orig(self, *args, **kwargs))

    orig = getattr(int, name)
    return functools.wraps(orig)(inner)


def _gen_tuple_proxy_method(name):
    def inner(self, *args, **kwargs):
        return tuple(map(self.__class__, orig(self, *args, **kwargs)))

    orig = getattr(int, name)
    return functools.wraps(orig)(inner)


class FixedWidthInt(int, lang.Abstract):
    BITS: ta.ClassVar[int]

    #

    class Mode(enum.StrEnum):
        CHECK = enum.auto()
        CLAMP = enum.auto()
        WRAP = enum.auto()

    MODE: ta.ClassVar[Mode]

    SIGNED: ta.ClassVar[bool]

    #

    MIN: ta.ClassVar[int]
    MAX: ta.ClassVar[int]

    MASK: ta.ClassVar[int]

    #

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        if lang.is_abstract_class(cls):
            return

        #

        bits = check.single({
            check.isinstance(bbits, int)
            for bcls in cls.__mro__
            if (bbits := bcls.__dict__.get('BITS')) is not None
        })

        #

        mode_base = _get_exclusive_base_cls(cls, _MODE_BASE_CLASSES)

        if mode_base is CheckedInt:
            cls.MODE = cls.Mode.CHECK

        elif mode_base is ClampedInt:
            cls.MODE = cls.Mode.CLAMP

        elif mode_base is WrappedInt:
            cls.MODE = cls.Mode.WRAP

        else:
            raise TypeError(cls)

        #

        sign_base = _get_exclusive_base_cls(cls, _SIGN_BASE_CLASSES)

        if sign_base is SignedInt:
            cls.SIGNED = True
            cls.MIN = -(1 << (bits - 1))
            cls.MAX = (1 << (bits - 1)) - 1

        elif sign_base is UnsignedInt:
            cls.SIGNED = False
            cls.MIN = 0
            cls.MAX = (1 << bits) - 1

        else:
            raise TypeError(cls)

        #

        cls.MASK = (1 << bits) - 1

    #

    @classmethod
    def clamp(cls, value: int) -> int:
        return max(min(value, cls.MAX), cls.MIN)

    @classmethod
    def check(cls, value: int) -> int:
        if value > cls.MAX:
            raise OverflowFixedWidthIntError(value)
        elif value < cls.MIN:
            raise UnderflowFixedWidthIntError(value)
        return value

    @classmethod
    def wrap(cls, value: int) -> int:
        return ((value - cls.MIN) & cls.MASK) + cls.MIN

    #

    _SCALAR_PROXY_METHODS = frozenset([
        '__abs__',
        '__add__',
        '__and__',
        '__floordiv__',
        '__invert__',
        '__lshift__',
        '__mod__',
        '__mul__',
        '__neg__',
        '__or__',
        '__pos__',
        '__pow__',
        '__radd__',
        '__rand__',
        '__rfloordiv__',
        '__rlshift__',
        '__rmod__',
        '__rmul__',
        '__ror__',
        '__rpow__',
        '__rrshift__',
        '__rshift__',
        '__rsub__',
        '__rtruediv__',
        '__rxor__',
        '__sub__',
        '__truediv__',
        '__xor__',
    ])

    _TUPLE_PROXY_METHODS = frozenset([
        '__divmod__',
        '__rdivmod__',
    ])

    for _proxy_name in _SCALAR_PROXY_METHODS:
        locals()[_proxy_name] = _gen_scalar_proxy_method(_proxy_name)
    for _proxy_name in _TUPLE_PROXY_METHODS:
        locals()[_proxy_name] = _gen_tuple_proxy_method(_proxy_name)
    del _proxy_name

    #

    def __invert__(self) -> ta.Self:
        if not self.SIGNED:
            return self.__class__(~int(self) & self.MASK)
        else:
            return self.__class__(super().__invert__())

    #

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({int(self)})'


##


class SignedInt(FixedWidthInt, lang.Abstract):
    pass


class UnsignedInt(FixedWidthInt, lang.Abstract):
    pass


_SIGN_BASE_CLASSES: tuple[type[FixedWidthInt], ...] = (
    SignedInt,
    UnsignedInt,
)


#


class CheckedInt(FixedWidthInt, lang.Abstract):
    def __new__(cls: type[FixedWidthIntT], value: int) -> FixedWidthIntT:
        return super().__new__(cls, cls.check(value))  # type: ignore[misc]


class ClampedInt(FixedWidthInt, lang.Abstract):
    def __new__(cls: type[FixedWidthIntT], value: int) -> FixedWidthIntT:
        return super().__new__(cls, cls.clamp(value))  # type: ignore[misc]


class WrappedInt(FixedWidthInt, lang.Abstract):
    def __new__(cls: type[FixedWidthIntT], value: int) -> FixedWidthIntT:
        return super().__new__(cls, cls.wrap(value))  # type: ignore[misc]


_MODE_BASE_CLASSES: tuple[type[FixedWidthInt], ...] = (
    CheckedInt,
    ClampedInt,
    WrappedInt,
)


#


class AnyInt8(FixedWidthInt, lang.Abstract):
    BITS = 8


class AnyInt16(FixedWidthInt, lang.Abstract):
    BITS = 16


class AnyInt32(FixedWidthInt, lang.Abstract):
    BITS = 32


class AnyInt64(FixedWidthInt, lang.Abstract):
    BITS = 64


class AnyInt128(FixedWidthInt, lang.Abstract):
    BITS = 128


##


class CheckedInt8(CheckedInt, SignedInt, AnyInt8):
    pass


class CheckedInt16(CheckedInt, SignedInt, AnyInt16):
    pass


class CheckedInt32(CheckedInt, SignedInt, AnyInt32):
    pass


class CheckedInt64(CheckedInt, SignedInt, AnyInt64):
    pass


class CheckedInt128(CheckedInt, SignedInt, AnyInt128):
    pass


#


class CheckedUint8(CheckedInt, UnsignedInt, AnyInt8):
    pass


class CheckedUint16(CheckedInt, UnsignedInt, AnyInt16):
    pass


class CheckedUint32(CheckedInt, UnsignedInt, AnyInt32):
    pass


class CheckedUint64(CheckedInt, UnsignedInt, AnyInt64):
    pass


class CheckedUint128(CheckedInt, UnsignedInt, AnyInt128):
    pass


#


class ClampedInt8(ClampedInt, SignedInt, AnyInt8):
    pass


class ClampedInt16(ClampedInt, SignedInt, AnyInt16):
    pass


class ClampedInt32(ClampedInt, SignedInt, AnyInt32):
    pass


class ClampedInt64(ClampedInt, SignedInt, AnyInt64):
    pass


class ClampedInt128(ClampedInt, SignedInt, AnyInt128):
    pass


#


class ClampedUint8(ClampedInt, UnsignedInt, AnyInt8):
    pass


class ClampedUint16(ClampedInt, UnsignedInt, AnyInt16):
    pass


class ClampedUint32(ClampedInt, UnsignedInt, AnyInt32):
    pass


class ClampedUint64(ClampedInt, UnsignedInt, AnyInt64):
    pass


class ClampedUint128(ClampedInt, UnsignedInt, AnyInt128):
    pass


#


class WrappedInt8(WrappedInt, SignedInt, AnyInt8):
    pass


class WrappedInt16(WrappedInt, SignedInt, AnyInt16):
    pass


class WrappedInt32(WrappedInt, SignedInt, AnyInt32):
    pass


class WrappedInt64(WrappedInt, SignedInt, AnyInt64):
    pass


class WrappedInt128(WrappedInt, SignedInt, AnyInt128):
    pass


#


class WrappedUint8(WrappedInt, UnsignedInt, AnyInt8):
    pass


class WrappedUint16(WrappedInt, UnsignedInt, AnyInt16):
    pass


class WrappedUint32(WrappedInt, UnsignedInt, AnyInt32):
    pass


class WrappedUint64(WrappedInt, UnsignedInt, AnyInt64):
    pass


class WrappedUint128(WrappedInt, UnsignedInt, AnyInt128):
    pass



# ruff: noqa: I001
import typing as _ta

from .. import lang as _lang


if _ta.TYPE_CHECKING:
    from . import api  # noqa
else:
    globals()['api'] = _lang.proxy_import('.api', __package__)

from .dbs import (  # noqa
    DbLoc,
    DbSpec,
    DbType,
    DbTypes,
    HostDbLoc,
    UrlDbLoc,
)

from .qualifiedname import (  # noqa
    QualifiedName,
    qn,
)

if _ta.TYPE_CHECKING:
    from . import queries  # noqa
else:
    globals()['queries'] = _lang.proxy_import('.queries', __package__)



import abc
import contextlib
import typing as ta

from .. import check
from .. import lang


V = ta.TypeVar('V')
MK = ta.TypeVar('MK')
MV = ta.TypeVar('MV')

SetMap: ta.TypeAlias = ta.Mapping[MK, ta.AbstractSet[MV]]


##


class DirectedGraph(ta.Generic[V], lang.Abstract):
    @abc.abstractmethod
    def get_successors(self, vertex: V) -> ta.Collection[V]:
        raise NotImplementedError

    @abc.abstractmethod
    def yield_depth_first(self, root: V) -> ta.Iterator[V]:
        raise NotImplementedError


class ListDictDirectedGraph(DirectedGraph[V]):
    def __init__(self, items: ta.Iterable[tuple[V, ta.Iterable[V]]]) -> None:
        super().__init__()

        lst_dct: dict[V, list[V]] = {}
        all_children = set()
        for parent, children in items:
            check.not_in(parent, lst_dct)
            lst = []
            seen = set()
            for child in children:
                if child not in seen:
                    seen.add(child)
                    lst.append(child)
                    all_children.add(child)
            lst_dct[parent] = lst
        check.empty(all_children - set(lst_dct))
        self._lst_dct = lst_dct

    def get_successors(self, vertex: V) -> ta.Collection[V]:
        return self._lst_dct[vertex]

    def yield_depth_first(self, root: V) -> ta.Iterator[V]:
        stack: list[V] = [root]
        seen: set[V] = set()
        while stack:
            cur = stack.pop()
            yield cur
            for child in self._lst_dct[cur]:
                if child not in seen:
                    seen.add(child)
                    stack.append(child)


class DominatorTree(ta.Generic[V]):
    def __init__(self, graph: DirectedGraph[V], root: V) -> None:
        super().__init__()

        self._graph = check.not_none(graph)
        self._root = check.not_none(root)
        check.not_none(self._graph.get_successors(root))
        self._dfs = _Dfs(graph, root)

    @lang.cached_property
    def immediate_dominators(self) -> ta.Mapping[V, V]:
        return _ImmediateDominanceComputer(self._dfs).immediate_dominators

    @lang.cached_property
    def dominator_tree(self) -> SetMap[V, V]:
        tree: dict[V, set[V]] = {}
        for node, dom in self.immediate_dominators.items():
            tree.setdefault(dom, set()).add(node)
        return tree

    @lang.cached_property
    def deep_dominated(self) -> SetMap[V, V]:
        seen: set[V] = set()
        ret: dict[V, set[V]] = {}

        def rec(node: V) -> ta.Collection[V]:
            check.not_in(node, seen)
            seen.add(node)
            # FIXME: pyrsistent
            st = set()
            for child in self.dominator_tree.get(node, []):
                st.add(child)
                st.update(rec(child))
            if st:
                ret[node] = st
            return st

        rec(self._root)
        return ret

    @lang.cached_property
    def dominance_frontiers(self) -> SetMap[V, V]:
        dominance_frontiers: dict[V, set[V]] = {}

        for x in self.reverse_topological_traversal:
            dfx = dominance_frontiers.setdefault(x, set())

            for y in self._graph.get_successors(x):
                if self.immediate_dominators[y] != x:
                    dfx.add(y)

            for z in self.dominator_tree.get(x, []):
                for y in dominance_frontiers.get(z, []):
                    if self.immediate_dominators[y] != x:
                        dfx.add(y)

        return {k: v for k, v in dominance_frontiers.items() if v}

    @lang.cached_property
    def topological_traversal(self) -> list[V]:
        # FIXME: LinkedList
        lst: list[V] = []

        for node in self._dfs.vertex:
            try:
                idx = lst.index(self.immediate_dominators[node])
            except (KeyError, ValueError):
                lst.append(node)
            else:
                lst.insert(idx + 1, node)

        return lst

    @lang.cached_property
    def reverse_topological_traversal(self) -> list[V]:
        return list(reversed(self.topological_traversal))


class _Dfs(ta.Generic[V]):
    def __init__(self, graph: DirectedGraph[V], root: V) -> None:
        super().__init__()

        semi: dict[V, int] = {}
        vertex: list[V] = []
        parent: dict[V, V] = {}
        pred: dict[V, set[V]] = {}
        label: dict[V, V] = {}

        for node in graph.yield_depth_first(root):
            if node not in semi:
                vertex.append(node)

                check.not_in(node, semi)
                semi[node] = len(semi)
                check.not_in(node, label)
                label[node] = node

                for child in graph.get_successors(node):
                    pred.setdefault(child, set()).add(node)
                    if child not in semi:
                        check.not_in(child, parent)
                        parent[child] = node

        self._semi = semi
        self._vertex = vertex
        self._parent = parent
        self._pred = pred
        self._label = label

    @property
    def semi(self) -> dict[V, int]:
        return self._semi

    @property
    def vertex(self) -> list[V]:
        return self._vertex

    @property
    def parent(self) -> dict[V, V]:
        return self._parent

    @property
    def pred(self) -> dict[V, set[V]]:
        return self._pred

    @property
    def label(self) -> dict[V, V]:
        return self._label


class _ImmediateDominanceComputer(ta.Generic[V]):
    def __init__(self, dfs: _Dfs[V]) -> None:
        super().__init__()

        self._dfs: _Dfs[V] = check.isinstance(dfs, _Dfs)

        self._ancestor: dict[V, V] = {}
        self._semi = dict(self._dfs.semi)
        self._label = dict(self._dfs.label)

    @lang.cached_property
    def immediate_dominators(self) -> ta.Mapping[V, V]:
        idom: dict[V, V] = {}
        bucket: dict[V, set[V]] = {}

        last_semi_number = len(self._semi) - 1

        for i in range(last_semi_number, 0, -1):
            w = self._dfs.vertex[i]
            p = self._dfs.parent[w]

            semidominator = self._semi[w]
            for v in self._dfs.pred.get(w, []):
                semidominator = min(semidominator, self._semi[self._eval(v)])

            self._semi[w] = semidominator
            bucket.setdefault(self._dfs.vertex[semidominator], set()).add(w)

            self._ancestor[w] = p

            for v in bucket.get(p, []):
                u = self._eval(v)

                if self._semi[u] < self._semi[v]:
                    idom[v] = u
                else:
                    idom[v] = p

            with contextlib.suppress(KeyError):
                del bucket[p]

        for i in range(1, last_semi_number + 1):
            w = self._dfs.vertex[i]

            if idom[w] != self._dfs.vertex[self._semi[w]]:
                idom[w] = idom[idom[w]]

        return idom

    def _eval(self, v: V) -> V:
        self._compress(v)
        return self._label[v]

    def _compress(self, v: V) -> None:
        worklist: list[V] = [v]

        a = self._ancestor.get(v)

        while a in self._ancestor:
            worklist.append(a)
            a = self._ancestor[a]

        ancestor = worklist.pop()
        least_semi = self._semi[self._label[ancestor]]

        while worklist:
            descendent = worklist.pop()
            current_semi = self._semi[self._label[descendent]]

            if current_semi > least_semi:
                self._label[descendent] = self._label[ancestor]
            else:
                least_semi = current_semi

            ancestor = descendent
